# Unit Tests for the Enum Combo Box

import pytest
from logging import ERROR

from qtpy.QtCore import Slot, Qt

from ...widgets.enum_combo_box import PyDMEnumComboBox
from ... import data_plugins


# --------------------
# POSITIVE TEST CASES
# --------------------

def test_construct(qtbot):
    """
    Test the construction of the widget.

    Expectations:
    All the default values are properly set.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    assert pydm_enumcombobox._has_enums is False
    assert pydm_enumcombobox.contextMenuPolicy() == Qt.DefaultContextMenu
    assert pydm_enumcombobox.contextMenuEvent == pydm_enumcombobox.open_context_menu


@pytest.mark.parametrize("enums", [
    ("spam", "eggs", "ham"),
    ("spam",),
    ("",),
])
def test_set_items(qtbot, enums):
    """
    Test the populating of enum string (choices) to the widget.

    Expectations:
    All enum strings are populated to the widget, and the _has_enum flag is set to True if the enum string list is not
    empty.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    enums : tuple
        A list of enum strings to be populated as choices to the widget.
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    assert pydm_enumcombobox.count() == 0
    pydm_enumcombobox.set_items(enums)
    assert pydm_enumcombobox.count() == len(enums)

    assert all([enums[i] == pydm_enumcombobox.itemText(i) for i in range(0, len(enums))])
    assert pydm_enumcombobox._has_enums is True if len(enums) else pydm_enumcombobox._has_enums is False


@pytest.mark.parametrize("connected, write_access, has_enum, is_app_read_only", [
    (True, True, True, True),
    (True, True, True, False),

    (True, True, False, True),
    (True, True, False, False),

    (True, False, False, True),
    (True, False, False, False),

    (True, False, True, True),
    (True, False, True, False),

    (False, True, True, True),
    (False, True, True, False),

    (False, False, True, True),
    (False, False, True, False),

    (False, True, False, True),
    (False, True, False, False),

    (False, False, False, True),
    (False, False, False, False),
])
def test_check_enable_state(qtbot, signals, connected, write_access, has_enum, is_app_read_only):
    """
    Test the tooltip generated depending on the channel connection, write access, whether the widget has enum strings,
    and whether the app is read-only.

    Expectations:
    1. If the data channel is disconnected, the widget's tooltip will display "PV is disconnected"
    2. If the data channel is connected, but it has no write access:
        a. If the app is read-only, the tooltip will read  "Running PyDM on Read-Only mode."
        b. If the app is not read-only, the tooltip will read "Access denied by Channel Access Security."
    3. If the widget does not have any enum strings, the tooltip will display "Enums not available".

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    connected : bool
        True if the channel is connected; False otherwise
    write_access : bool
        True if the widget has write access to the channel; False otherwise
    has_enum: bool
        True if the widget has enum strings populated; False if the widget contains no enum strings (empty of choices)
    is_app_read_only : bool
        True if the PyDM app is read-only; False otherwise
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    signals.write_access_signal[bool].connect(pydm_enumcombobox.writeAccessChanged)
    signals.write_access_signal[bool].emit(write_access)

    signals.connection_state_signal[bool].connect(pydm_enumcombobox.connectionStateChanged)
    signals.connection_state_signal[bool].emit(connected)

    if has_enum:
        signals.enum_strings_signal[tuple].connect(pydm_enumcombobox.enumStringsChanged)
        signals.enum_strings_signal[tuple].emit(("START", "STOP", "PAUSE"))
        assert pydm_enumcombobox._has_enums

    data_plugins.set_read_only(is_app_read_only)

    original_tooltip = "Original Tooltip"
    pydm_enumcombobox.setToolTip(original_tooltip)
    pydm_enumcombobox.check_enable_state()

    actual_tooltip = pydm_enumcombobox.toolTip()
    if not pydm_enumcombobox._connected:
        assert "PV is disconnected." in actual_tooltip
    elif not write_access:
        if data_plugins.is_read_only():
            assert "Running PyDM on Read-Only mode." in actual_tooltip
        else:
            assert "Access denied by Channel Access Security." in actual_tooltip
    elif not pydm_enumcombobox._has_enums:
        assert "Enums not available" in actual_tooltip


@pytest.mark.parametrize("values, selected_index, expected", [
    (("RUN", "STOP"), 0, "RUN"),
    (("RUN", "STOP"), 1, "STOP"),
    (("RUN", "STOP"), "RUN", "RUN"),
    (("RUN", "STOP"), "STOP", "STOP"),
])
def test_enum_strings_changed(qtbot, signals, values, selected_index, expected):
    """
    Test the widget's handling of enum strings, which are choices presented to the user, and the widget's ability to
    update the selected enum string when the user provides a choice index.

    This test will also cover value_changed() testing.

    Expectations:
    The widget displays the correct enum string whose index from the enum string tuple is selected by the user.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    values : tuple
        A set of enum strings for the user to choose from
    selected_index : int
        The index from the enum string tuple chosen by the user
    expected : str
        The expected enum string displayed by the widget after receiving the user's choice index
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    signals.enum_strings_signal.connect(pydm_enumcombobox.enumStringsChanged)
    signals.enum_strings_signal.emit(values)

    signals.new_value_signal[type(selected_index)].connect(pydm_enumcombobox.channelValueChanged)
    signals.new_value_signal[type(selected_index)].emit(selected_index)

    assert pydm_enumcombobox.value == selected_index
    assert pydm_enumcombobox.currentText() == expected


@pytest.mark.parametrize("index", [
    0,
    1,
    -1,
])
def test_internal_combo_box_activated_int(qtbot, signals, index):
    """
    Test the the capability of the widget's activated slot in sending out a new enum string index value.

    Expectations:
    The value sent out from the "activated" slot reflects the correct new index value.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    index : int
        The new enum string index value
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    # Connect the send_value_signal also to the conftest's receiveValue slot to intercept and verify that the correct
    # new index value is sent out
    pydm_enumcombobox.send_value_signal[int].connect(signals.receiveValue)
    pydm_enumcombobox.activated[int].emit(index)

    assert signals.value == index


# --------------------
# NEGATIVE TEST CASES
# --------------------

@pytest.mark.parametrize("enums, expected_error_message", [
    (None, "Invalid enum value '{0}'. The value is expected to be a valid list of string values.".format(None)),
    ((None, "abc"), "Invalid enum type '{0}'. The expected type is 'string'.".format(type(None))),
    ((None, 123.456), "Invalid enum type '{0}'. The expected type is 'string'".format(type(None))),
    ((None, None, None), "Invalid enum type '{0}'. The expected type is 'string'".format(type(None))),
    ((123,),  "Invalid enum type '{0}'. The expected type is 'string'".format(type(123))),
    ((123.45,),  "Invalid enum type '{0}'. The expected type is 'string'".format(type(123.45))),
    ((123, 456), "Invalid enum type '{0}'. The expected type is 'string'".format(type(123))),
    ((123.456, None), "Invalid enum type '{0}'. The expected type is 'string'".format(type(123.456))),
    (("spam", 123, "eggs", "ham"), "Invalid enum type '{0}'. The expected type is 'string'".format(type(123))),
])
def test_set_items_neg(qtbot, caplog, enums, expected_error_message):
    """
    Test sending setting the widget with an undefined list of enum strings.

    Expectations:
    The correct error message is logged.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    caplog : fixture
        To capture the log messages
    enums : tuple
        A list of strings as enum strings (choices) to populate to the widget.
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    pydm_enumcombobox.set_items(enums)

    for record in caplog.records:
        assert record.levelno == ERROR
    assert expected_error_message in caplog.text


@pytest.mark.parametrize("values, selected_index, expected", [
    (("ON", "OFF"), 3, ""),
    (("ON", "OFF"), -1, ""),
])
def test_enum_strings_changed_incorrect_index(qtbot, signals, values, selected_index, expected):
    """
    Test the widget's handling of incorrectly provided enum string index.

    Expectations:
    The widget will display an empty string.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    value : tuple
        A set of enum strings for the user to choose from
    selected_index : int
        The incorrect (out-of-bound) index from the enum string tuple chosen by the user
    expected : int
        The expected text displayed by the widget to notify the user of the incorrect choice index
    """
    pydm_enumcombobox = PyDMEnumComboBox()
    qtbot.addWidget(pydm_enumcombobox)

    signals.enum_strings_signal.connect(pydm_enumcombobox.enumStringsChanged)
    signals.enum_strings_signal.emit(values)

    signals.new_value_signal[type(selected_index)].connect(pydm_enumcombobox.channelValueChanged)
    signals.new_value_signal[type(selected_index)].emit(selected_index)

    assert pydm_enumcombobox.value == selected_index
    assert pydm_enumcombobox.currentText() == expected



from qtpy.QtCore import QModelIndex, QVariant
from .baseplot_table_model import BasePlotCurvesModel
from .baseplot_curve_editor import BasePlotCurveEditorDialog, PlotStyleColumnDelegate, RedrawModeColumnDelegate


class PyDMEventPlotCurvesModel(BasePlotCurvesModel):
    """ This is the data model used by the waveform plot curve editor.
    It basically acts as a go-between for the curves in a plot, and
    QTableView items.
    """

    def __init__(self, plot, parent=None):
        super(PyDMEventPlotCurvesModel, self).__init__(plot, parent=parent)
        self._column_names = ('Channel', 'Y Index', 'X Index') + self._column_names
        self._column_names += ('Buffer Size', 'Buffer Size Channel')

    def get_data(self, column_name, curve):
        if column_name == "Channel":
            if curve.address is None:
                return QVariant()
            return str(curve.address)
        elif column_name == "Y Index":
            if curve.y_idx is None:
                return QVariant()
            return curve.y_idx
        elif column_name == "X Index":
            if curve.x_idx is None:
                return QVariant()
            return curve.x_idx
        elif column_name == "Buffer Size":
            return curve.getBufferSize()
        elif column_name == "Buffer Size Channel":
            return curve.bufferSizeChannelAddress or ""
        return super(PyDMEventPlotCurvesModel, self).get_data(
            column_name, curve)

    def set_data(self, column_name, curve, value):
        if column_name == "Channel":
            curve.address = str(value)
        elif column_name == "Y Index":
            curve.y_idx = value
        elif column_name == "X Index":
            curve.x_idx = value
        elif column_name == "Buffer Size":
            curve.setBufferSize(int(value))
        elif column_name == "Buffer Size Channel":
            if len(str(value).strip()) < 1:
                value = None
            curve.bufferSizeChannelAddress = str(value)
        else:
            return super(PyDMEventPlotCurvesModel, self).set_data(
                column_name=column_name, curve=curve, value=value)
        return True

    def append(self, channel=None, y_idx=None, x_idx=None, name=None, color=None):
        self.beginInsertRows(QModelIndex(), len(self._plot._curves),
                             len(self._plot._curves))
        self._plot.addChannel(channel, y_idx, x_idx, name, color)
        self.endInsertRows()

    def removeAtIndex(self, index):
        self.beginRemoveRows(QModelIndex(), index.row(), index.row())
        self._plot.removeChannelAtIndex(index.row())
        self.endRemoveRows()


class EventPlotCurveEditorDialog(BasePlotCurveEditorDialog):
    """EventPlotCurveEditorDialog is a QDialog that is used in Qt Designer
    to edit the properties of the curves in a waveform plot.  This dialog is
    shown when you double-click the plot, or when you right click it and
    choose 'edit curves'.

    This thing is mostly just a wrapper for a table view, with a couple
    buttons to add and remove curves, and a button to save the changes."""
    TABLE_MODEL_CLASS = PyDMEventPlotCurvesModel

    def __init__(self, plot, parent=None):
        super(EventPlotCurveEditorDialog, self).__init__(plot, parent)

        plot_style_delegate = PlotStyleColumnDelegate(self, self.table_model, self.table_view)
        plot_style_delegate.hideColumns(hide_line_columns=False, hide_bar_columns=True)



from qtpy.QtWidgets import QTableView
from ...widgets.baseplot import BasePlot
from ...widgets.baseplot_curve_editor import (AxisColumnDelegate, ColorColumnDelegate, LineColumnDelegate,
                                              SymbolColumnDelegate, RedrawModeColumnDelegate, PlotStyleColumnDelegate)
from ...widgets.baseplot_table_model import BasePlotCurvesModel
from ...widgets.scatterplot_curve_editor import ScatterPlotCurveEditorDialog
from ...widgets.timeplot_curve_editor import TimePlotCurveEditorDialog
from ...widgets.waveformplot import WaveformCurveItem
from ...widgets.waveformplot_curve_editor import WaveformPlotCurveEditorDialog


def test_waveform_curve_editor(qtbot):
    """
    Ensure that the waveform curve editor looks and functions as expected
    """

    # Create waveform plot curve editor along with its associated plot. Ensure it shows.
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)

    curve_editor = WaveformPlotCurveEditorDialog(base_plot)
    qtbot.addWidget(curve_editor)
    curve_editor.show()

    table_model = curve_editor.table_model
    table_view = curve_editor.table_view

    # Verify that the drop downs for columns with non built-in types are all put in the correct place
    # Note: We do need to check these on each individual type of curve editor (see below tests) and not just
    # in the base plot editor since each plot type can have varying numbers of columns
    color_index = table_model.getColumnIndex('Color')
    line_style_index = table_model.getColumnIndex('Line Style')
    symbol_index = table_model.getColumnIndex('Symbol')
    redraw_mode_index = table_model.getColumnIndex('Redraw Mode')
    plot_style_index = table_model.getColumnIndex('Style')

    assert type(table_view.itemDelegateForColumn(color_index)) is ColorColumnDelegate
    assert type(table_view.itemDelegateForColumn(line_style_index)) is LineColumnDelegate
    assert type(table_view.itemDelegateForColumn(symbol_index)) is SymbolColumnDelegate
    assert type(table_view.itemDelegateForColumn(redraw_mode_index)) is RedrawModeColumnDelegate
    assert type(table_view.itemDelegateForColumn(plot_style_index)) is PlotStyleColumnDelegate


def test_timeplot_curve_editor(qtbot):
    """
    Ensure that the time plot curve editor looks and functions as expected
    """

    # Create time plot curve editor along with its associated plot. Ensure it shows.
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)

    curve_editor = TimePlotCurveEditorDialog(base_plot)
    qtbot.addWidget(curve_editor)
    curve_editor.show()

    table_model = curve_editor.table_model
    table_view = curve_editor.table_view

    # Verify that the drop downs for columns with non built-in types are all put in the correct place
    color_index = table_model.getColumnIndex('Color')
    line_style_index = table_model.getColumnIndex('Line Style')
    symbol_index = table_model.getColumnIndex('Symbol')
    plot_style_index = table_model.getColumnIndex('Style')

    assert type(table_view.itemDelegateForColumn(color_index)) is ColorColumnDelegate
    assert type(table_view.itemDelegateForColumn(line_style_index)) is LineColumnDelegate
    assert type(table_view.itemDelegateForColumn(symbol_index)) is SymbolColumnDelegate
    assert type(table_view.itemDelegateForColumn(plot_style_index)) is PlotStyleColumnDelegate


def test_scatterplot_editor(qtbot):
    """
    Ensure that the scatter plot curve editor looks and functions as expected
    """

    # Create scatter plot curve editor along with its associated plot. Ensure it shows.
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)

    curve_editor = ScatterPlotCurveEditorDialog(base_plot)
    qtbot.addWidget(curve_editor)
    curve_editor.show()

    table_model = curve_editor.table_model
    table_view = curve_editor.table_view

    # Verify that the drop downs for columns with non built-in types are all put in the correct place
    color_index = table_model.getColumnIndex('Color')
    line_style_index = table_model.getColumnIndex('Line Style')
    symbol_index = table_model.getColumnIndex('Symbol')
    redraw_mode_index = table_model.getColumnIndex('Redraw Mode')

    assert type(table_view.itemDelegateForColumn(color_index)) is ColorColumnDelegate
    assert type(table_view.itemDelegateForColumn(line_style_index)) is LineColumnDelegate
    assert type(table_view.itemDelegateForColumn(symbol_index)) is SymbolColumnDelegate
    assert type(table_view.itemDelegateForColumn(redraw_mode_index)) is RedrawModeColumnDelegate


def test_axis_editor(qtbot):
    """
    Ensure that the axis editor tab in the curve editor looks and functions as expected
    """

    base_plot = BasePlot()
    qtbot.addWidget(base_plot)
    curve_editor = WaveformPlotCurveEditorDialog(base_plot)

    axis_model = curve_editor.axis_model
    axis_view = curve_editor.axis_view

    # Verify the column count is correct, and the axis column delegate is placed correctly
    axis_orientation_index = axis_model._column_names.index('Y-Axis Orientation')
    assert type(axis_view.itemDelegateForColumn(axis_orientation_index)) is AxisColumnDelegate


def test_plot_style_column_delegate(qtbot):
    """ Verify the functionality of the show/hide column feature """

    # Set up a plot with three data items. Two will be plotted as lines, and one as bars.
    base_plot = BasePlot()
    qtbot.addWidget(base_plot)
    line_item_1 = WaveformCurveItem()
    line_item_2 = WaveformCurveItem()
    bar_item = WaveformCurveItem(plot_style='Bar')
    plot_curves_model = BasePlotCurvesModel(plot=base_plot)
    table_view = QTableView()
    table_view.setModel(plot_curves_model)
    plot_style_column_delegate = PlotStyleColumnDelegate(parent=base_plot,
                                                         table_model=plot_curves_model,
                                                         table_view=table_view)

    base_plot.addCurve(line_item_1)
    plot_style_column_delegate.toggleColumnVisibility()

    # With only the line style curve displayed the four line columns should be shown
    for column in plot_style_column_delegate.line_columns_to_toggle:
        assert not table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))
    # And the four bar columns should be hidden
    for column in plot_style_column_delegate.bar_columns_to_toggle:
        assert table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))

    # Now add an additional line curve and a bar curve. All 8 columns should now be visible since it's a mixed plot
    base_plot.addCurve(line_item_2)
    base_plot.addCurve(bar_item)

    plot_style_column_delegate.toggleColumnVisibility()
    for column in plot_style_column_delegate.line_columns_to_toggle:
        assert not table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))
    for column in plot_style_column_delegate.bar_columns_to_toggle:
        assert not table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))

    # Finally remove both line curves to test the last option, nothing but bar style curves. All line options should
    # be hidden, while the bar options should still be shown
    base_plot.removeCurve(line_item_1)
    base_plot.removeCurve(line_item_2)

    plot_style_column_delegate.toggleColumnVisibility()
    for column in plot_style_column_delegate.line_columns_to_toggle:
        assert table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))
    for column in plot_style_column_delegate.bar_columns_to_toggle:
        assert not table_view.isColumnHidden(plot_curves_model.getColumnIndex(column))



import pytest
import logging
import numpy as np
from collections import OrderedDict
from pyqtgraph import AxisItem, BarGraphItem
from unittest import mock
from ...widgets.channel import PyDMChannel
from ...widgets.timeplot import TimePlotCurveItem, PyDMTimePlot, TimeAxisItem, MINIMUM_BUFFER_SIZE, DEFAULT_BUFFER_SIZE
from ...utilities import remove_protocol

logger = logging.getLogger(__name__)


@pytest.mark.parametrize("channel_address, name", [
    ("ca://test_value:Float", "test_name"),
    ("ca://test_value:Float", ""),
    ("ca://test_value:Float", None),
    ("", None),
    (None, None)
])
def test_timeplotcurveitem_construct(qtbot, channel_address, name):
    pydm_timeplot_curve_item = TimePlotCurveItem(channel_address=channel_address, name=name)

    if not name:
        assert pydm_timeplot_curve_item.to_dict()["name"] == remove_protocol(channel_address) if channel_address else \
            not pydm_timeplot_curve_item.to_dict()["name"]

    assert pydm_timeplot_curve_item._bufferSize == MINIMUM_BUFFER_SIZE
    assert pydm_timeplot_curve_item._update_mode == PyDMTimePlot.OnValueChange
    assert np.array_equal(pydm_timeplot_curve_item.data_buffer, np.zeros((2, pydm_timeplot_curve_item._bufferSize),
                                                                         order='f', dtype=float))
    assert pydm_timeplot_curve_item.connected is False
    assert pydm_timeplot_curve_item.points_accumulated == 0
    assert pydm_timeplot_curve_item.latest_value is None
    assert pydm_timeplot_curve_item.address == channel_address if channel_address else \
        pydm_timeplot_curve_item.address is None


@pytest.mark.parametrize("channel_address, name", [
    ("ca://test_value:Float", "test_name"),
    ("ca://test_value:Float", ""),
    ("ca://test_value:Float", None),
    ("", None),
    (None, None)
])
def test_timeplotcurveitem_to_dict(qtbot, channel_address, name):
    pydm_timeplot_curve_item = TimePlotCurveItem(channel_address=channel_address, name=name)

    dictionary = pydm_timeplot_curve_item.to_dict()
    assert isinstance(dictionary, OrderedDict)

    assert dictionary["channel"] == channel_address if channel_address else dictionary["channel"] is None
    if name:
        assert(dictionary["name"] == name)
    else:
        assert pydm_timeplot_curve_item.to_dict()["name"] == remove_protocol(channel_address) if channel_address else \
            not pydm_timeplot_curve_item.to_dict()["name"]


@pytest.mark.parametrize("new_address", [
    "new_address",
    "",
    None
])
def test_timeplotcurveitem_properties_and_setters(qtbot, new_address):
    pydm_timeplot_curve_item = TimePlotCurveItem()

    assert pydm_timeplot_curve_item.address is None

    pydm_timeplot_curve_item.address = new_address
    if new_address:
        assert isinstance(pydm_timeplot_curve_item.channel, PyDMChannel)
        assert pydm_timeplot_curve_item.channel.address == new_address
    else:
        assert pydm_timeplot_curve_item.channel is None


def test_timeplotcurveitem_connection_state_changed(qtbot, signals):
    pydm_timeplot_curve_item = TimePlotCurveItem()
    assert pydm_timeplot_curve_item.connected is False

    signals.connection_state_signal.connect(pydm_timeplot_curve_item.connectionStateChanged)
    signals.connection_state_signal.emit(True)
    assert pydm_timeplot_curve_item.connected


@pytest.mark.parametrize("async_update, new_data", [
    (False, -10),
    (False, 10.2333),
    (True, 100),
    (True, -123.456)
])
def test_timeplotcurveitem_receive_value(qtbot, signals, async_update, new_data):
    """
    Also testing setUpdatesAsynchronously, resetUpdatesAsynchronously, and initialize_buffer
    Parameters
    ----------
    qtbot
    async_update

    Returns
    -------

    """
    pydm_timeplot_curve_item = TimePlotCurveItem()

    assert pydm_timeplot_curve_item._update_mode == PyDMTimePlot.OnValueChange

    pydm_timeplot_curve_item.setUpdatesAsynchronously(async_update)
    if async_update:
        assert pydm_timeplot_curve_item._update_mode == PyDMTimePlot.AtFixedRated if async_update else \
            pydm_timeplot_curve_item._update_mode == PyDMTimePlot.OnValueChange

    expected_data_buffer = np.zeros((2, pydm_timeplot_curve_item._bufferSize), order='f', dtype=float)
    expected_data_buffer[0] = pydm_timeplot_curve_item.data_buffer[0]
    assert np.array_equal(expected_data_buffer, pydm_timeplot_curve_item.data_buffer)

    signals.new_value_signal[type(new_data)].connect(pydm_timeplot_curve_item.receiveNewValue)
    signals.new_value_signal[type(new_data)].emit(new_data)

    if async_update:
        assert np.array_equal(pydm_timeplot_curve_item.latest_value, new_data)
    else:
        assert np.array_equal(pydm_timeplot_curve_item.data_buffer[1, pydm_timeplot_curve_item._bufferSize - 1],
                              new_data)
        assert pydm_timeplot_curve_item.points_accumulated == 1

    pydm_timeplot_curve_item.resetUpdatesAsynchronously()
    assert pydm_timeplot_curve_item._update_mode == PyDMTimePlot.OnValueChange


@pytest.mark.parametrize("async_update, new_data", [
    (False, -10),
    (False, 10.2333),
    (True, 100),
    (True, -123.456)
])
def test_timeplotcurveitem_async_update(qtbot, signals, async_update, new_data):
    pydm_timeplot_curve_item = TimePlotCurveItem()

    assert pydm_timeplot_curve_item._update_mode == PyDMTimePlot.OnValueChange

    pydm_timeplot_curve_item.setUpdatesAsynchronously(async_update)

    signals.new_value_signal[type(new_data)].connect(pydm_timeplot_curve_item.receiveNewValue)
    signals.new_value_signal[type(new_data)].emit(new_data)

    signals.new_value_signal[type(new_data)].connect(pydm_timeplot_curve_item.asyncUpdate)
    signals.new_value_signal[type(new_data)].emit(new_data)

    if async_update:
        assert np.array_equal(pydm_timeplot_curve_item.data_buffer[1, pydm_timeplot_curve_item._bufferSize - 1],
                              new_data)
        assert pydm_timeplot_curve_item.points_accumulated == 1
    else:
        assert pydm_timeplot_curve_item.points_accumulated == 2


def test_timeplotcurve_initialize_buffer(qtbot):
    pydm_timeplot_curve_item = TimePlotCurveItem()

    pydm_timeplot_curve_item.initialize_buffer()

    assert pydm_timeplot_curve_item.points_accumulated == 0
    expected_data_buffer = np.zeros((2, pydm_timeplot_curve_item._bufferSize), order='f', dtype=float)
    expected_data_buffer[0] = pydm_timeplot_curve_item.data_buffer[0]

    assert np.array_equal(expected_data_buffer, pydm_timeplot_curve_item.data_buffer)


@pytest.mark.parametrize("new_buffer_size, expected_set_buffer_size", [
    (0, MINIMUM_BUFFER_SIZE),
    (-5, MINIMUM_BUFFER_SIZE),
    (100, 100),
    (MINIMUM_BUFFER_SIZE + 1, MINIMUM_BUFFER_SIZE + 1)
])
def test_timeplotcurve_get_set_reset_buffer_size(qtbot, new_buffer_size, expected_set_buffer_size):
    pydm_timeplot_curve_item = TimePlotCurveItem()

    assert pydm_timeplot_curve_item.getBufferSize() == MINIMUM_BUFFER_SIZE

    pydm_timeplot_curve_item.setBufferSize(new_buffer_size)
    assert pydm_timeplot_curve_item.getBufferSize() == expected_set_buffer_size

    pydm_timeplot_curve_item.resetBufferSize()
    assert pydm_timeplot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE


def test_timeplotcurve_max_x(qtbot):
    pydm_timeplot_curve_item = TimePlotCurveItem()

    pydm_timeplot_curve_item.data_buffer[0, pydm_timeplot_curve_item._bufferSize - 1] = -1
    pydm_timeplot_curve_item.data_buffer[1, pydm_timeplot_curve_item._bufferSize - 1] = 100

    max_value = pydm_timeplot_curve_item.max_x()
    assert max_value == pydm_timeplot_curve_item.data_buffer[0, pydm_timeplot_curve_item._bufferSize - 1]
    assert pydm_timeplot_curve_item.data_buffer[1, pydm_timeplot_curve_item._bufferSize - 1] == 100


def test_timeaxisitem_tickstrings():
    time_axis_item = TimeAxisItem("bottom")
    assert len(time_axis_item.tickStrings((10, 20, 30), 1, 1)) > 0


def test_pydmtimeplot_construct(qtbot):
    pydm_timeplot = PyDMTimePlot()
    qtbot.addWidget(pydm_timeplot)

    assert isinstance(pydm_timeplot._bottom_axis, AxisItem)
    assert pydm_timeplot._bottom_axis.orientation == "bottom"


@mock.patch('pydm.widgets.timeplot.TimePlotCurveItem.setData')
@mock.patch('pyqtgraph.BarGraphItem.setOpts')
def test_redraw_plot(mocked_set_opts, mocked_set_data, qtbot, monkeypatch):
    """ Test redrawing a time plot using both a line and a bar graph """

    # Create a time plot and add two data items to it, one to be rendered as a line and one as a bar graph
    time_plot = PyDMTimePlot()
    line_item = TimePlotCurveItem()
    bar_item = TimePlotCurveItem(plot_style='Bar')
    bar_item.bar_graph_item = BarGraphItem(x=[], height=[], width=1.0)
    time_plot.addCurve(line_item)
    time_plot.addCurve(bar_item)

    # Setup some mock data for our data items
    line_item.data_buffer = np.array([[1, 5, 10], [10, 15, 12]], dtype=float)
    line_item.points_accumulated = 3
    bar_item.data_buffer = np.array([[0.5, 1, 1.5, 2, 10, 11], [45, 50, 52, 40, 24, 30]], dtype=float)
    bar_item.points_accumulated = 6

    time_plot.set_needs_redraw()
    time_plot.plotItem.setXRange(1, 10)  # Sets the visible x-range of this time plot

    monkeypatch.setattr(time_plot, 'updateXAxis', lambda: None)  # Ensure the view box range is deterministic

    # Simulate a redraw of the plot
    time_plot.redrawPlot()

    # The line item should result in a call to set data displaying all available data points as defined above
    assert np.array_equal(mocked_set_data.call_args_list[2][1]['x'], np.array([1, 5, 10]))
    assert np.array_equal(mocked_set_data.call_args_list[2][1]['y'], np.array([10, 15, 12]))

    # Because we want to limit rendering of bar graphs to only those data points which are visible, we should
    # see a call made with only 4 of the 6 data points. The data points associated with the x values 0.5 and 11
    # were omitted since they fell outside the viewable range of 1 to 10.
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]['x'], np.array([1, 1.5, 2, 10]))
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]['height'], np.array([50, 52, 40, 24]))

    # After a call to redraw, the plot returns to this state until more data arrives
    assert not time_plot._needs_redraw



import os
import pytest
from pydm import Display
from pydm.display import load_file, load_py_file, _compile_ui_file, _load_compiled_ui_into_display, ScreenTarget
from qtpy.QtWidgets import QLabel

# The path to the .ui file used in these tests
test_ui_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_data", "test.ui")

test_ui_with_macros_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_data", "macro_test.ui"
)

# The path to the .py files used in these tests
no_display_test_py_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_data", "no_display_test_file.py")

valid_display_test_py_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_data", "valid_display_test_file.py")


def test_ui_filename_arg(qtbot):
    """If you supply a valid filename argument, you shouldn't get any exceptions."""
    my_display = Display(parent=None, ui_filename=test_ui_path)
    qtbot.addWidget(my_display)


def test_reimplemented_ui_filename(qtbot):
    """If you reimplement ui_filename and return a valid filename, you
    shouldn't get any exceptions."""
    class TestDisplay(Display):
        def ui_filename(self):
            return test_ui_path
    my_display = TestDisplay(parent=None)
    qtbot.addWidget(my_display)


def test_nonexistant_ui_file_raises(qtbot):
    with pytest.raises(IOError):
        my_display = Display(parent=None, ui_filename="this_doesnt_exist.ui")

    class TestDisplay(Display):
        def ui_filename(self):
            return "this_doesnt_exist.ui"

    with pytest.raises(IOError):
        my_display = TestDisplay(parent=None)


def test_nonexistent_py_file_raises():
    """ Load a python file that does not exist and confirm the error raised is as expected """
    with pytest.raises(FileNotFoundError):
        load_py_file('this_doesnt_exist.py')


def test_doesnt_inherit_display_raises():
    """ Load a python file that does not inherit from PyDM Display and confirm the error raised is as expected """
    with pytest.raises(ValueError) as error_info:
        load_py_file(no_display_test_py_path)
    assert 'no class inheriting from Display' in str(error_info.value)


def test_load_valid_python_display_file(qtbot):
    """ Verify that loading a valid python only file inheriting from Display works as expected """
    display = load_py_file(valid_display_test_py_path)
    qtbot.addWidget(display)

    # Confirm that the file loaded everything as expected
    assert display.loaded_file() == valid_display_test_py_path
    assert display.ui_filename() == 'test.ui'
    assert display.macros() == {}
    assert display.previous_display is None
    assert display.next_display is None


def test_load_python_file_with_macros(qtbot):
    """ Attempt to add macros to the display while loading the file """
    macros = {'MACRO_1': 7, 'MACRO_2': 'test_string'}
    display = load_py_file(valid_display_test_py_path, macros=macros)
    qtbot.addWidget(display)
    assert display.loaded_file() == valid_display_test_py_path
    assert display.ui_filename() == 'test.ui'
    assert display.macros() == {'MACRO_1': 7, 'MACRO_2': 'test_string'}


def test_file_path_in_stylesheet_property(qtbot):
    """If you supply a valid filename argument, you shouldn't get any exceptions."""
    my_display = Display(parent=None, ui_filename=test_ui_path)
    qtbot.addWidget(my_display)
    my_display.setStyleSheet("test_stylesheet.css")
    test_css_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test_data", "test_stylesheet.css")
    with open(test_css_path) as css_file:
        css = css_file.read()
        # Assert that the stylesheet property is populated with the contents of the file.
        assert my_display.styleSheet() == css


def test_stylesheet_property_without_path(qtbot):
    """If you supply a valid filename argument, you shouldn't get any exceptions."""
    my_display = Display(parent=None, ui_filename=test_ui_path)
    qtbot.addWidget(my_display)
    css = "PyDMLabel { font-weight: bold; }"
    my_display.setStyleSheet(css)
    assert my_display.styleSheet() == css


def test_compile_ui_file():
    """
    Does a compile of our test ui file with uic and verifies the correct class is created,
    and the expected methods are added
    """
    code_string, class_name = _compile_ui_file(test_ui_path)
    assert class_name == 'Ui_Form'
    assert 'setupUi(self' in code_string
    assert 'retranslateUi(self' in code_string


def test_load_file_with_macros(qtbot):
    """
    Compiles and loads a ui file containing macros to verify there are no problems. Tests both an individual string
    and a list of strings.
    """
    try:
        # Rather than messing around with adding custom qt widgets just for one test, pretend like a QLabel
        # has a function that sets a value from a string list to verify its correctness
        commands_from_macro = []

        def setCommands(self, commands):
            """ Store what the commands were after macro parsing """
            nonlocal commands_from_macro
            commands_from_macro = commands
        QLabel.setCommands = setCommands

        # Compile the ui file into python code
        macros = {"test_label": "magnet_list",
                  "test_command": "grep -i 'string with spaces'",
                  "test_command_2": "echo hello",
                  "channel_with_dec_option": '.{"dec":{"n": 25}}'}
        code_string, class_name = _compile_ui_file(test_ui_with_macros_path)
        assert class_name == 'Ui_Form'

        # Parse and replace macros, then load into the display
        test_display = Display(macros=macros)
        _load_compiled_ui_into_display(code_string, class_name, test_display, macros)

        # Verify that the macros were replaced correctly
        assert test_display.ui.myLabel.text() == "magnet_list"
        assert test_display.ui.doubleQuotedLabel.text() == '.{"dec":{"n": 25}}'
        assert commands_from_macro == ["grep -i 'string with spaces'", "echo hello"]

    finally:
        del QLabel.setCommands


def test_load_file_with_help_display(qtbot):
    """
    Ensure that when a file containing help information is placed in the same directory as the display to load,
    that help display is loaded and available to the user. This test depends on a test.txt or test.html file
    being present in the same location as the file at test_ui_path.
    """
    test_display = load_file(test_ui_path, target=ScreenTarget.HOME)
    assert test_display.help_window is not None
    assert test_display.help_window.display_content.toPlainText() == 'This is a test help file for the test.ui display\n'



from qtpy import QtWidgets, QtGui, QtCore


def install_connection_inspector(parent, keys=None):
    """
    Install a QShortcut at the application which opens the PyDM Connection
    Inspector

    Parameters
    ----------
    parent : QWidget
        A shortcut is "listened for" by Qt's event loop when the shortcut's
        parent widget is receiving events.
    keys : QKeySequence, optional
        Default value is `Alt+C`
    """
    from pydm.connection_inspector import ConnectionInspector

    def show_inspector():
        c = ConnectionInspector(parent=parent)
        c.show()

    parent = parent or QtWidgets.QApplication.desktop()

    if keys is None:
        keys = QtGui.QKeySequence(QtCore.Qt.ALT + QtCore.Qt.Key_C)
    shortcut = QtWidgets.QShortcut(keys, parent);
    shortcut.setContext(QtCore.Qt.ApplicationShortcut)
    shortcut.activated.connect(show_inspector)



"""
iconfont provides a barebones system to get QIcons from icon fonts (like Font Awesome).
The inspiration and methodology come from the 'QtAwesome' module, which does exactly this, but
is a little too big, complicated, and flexible for PyDM's needs.
"""
import json
import os
import sys
from typing import Optional

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import QPoint, QRect, Qt, qRound
from qtpy.QtGui import (QColor, QFont, QFontDatabase, QIcon, QIconEngine,
                        QPainter, QPixmap)

if sys.version_info[0] == 3:
    unichr = chr


class IconFont(object):
    """IconFont represents an icon font.  Users will generally want
    to use IconFont.icon() to get a QIcon object for the character they want."""
    __instance = None

    def __init__(self):
        if self.__initialized:
            return
        self.font_file = "fontawesome.otf"  # specify these relative to this file.
        self.charmap_file = "fontawesome-charmap.json"
        self.font_name = None
        self.char_map = {}
        self.loaded_fonts = {}
        self.__initialized = True

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = object.__new__(IconFont)
            cls.__instance.__initialized = False
        return cls.__instance

    def load_font(self, ttf_filename, charmap_filename):
        """
        Load font from ``ttf_filename`` with a mapping defined in
        ``charmap_filename``.
        """
        def hook(obj):
            result = {}
            for key in obj:
                result[key] = unichr(int(obj[key], 16))
            return result

        if self.char_map:
            return

        cache_key = ttf_filename + "|" + charmap_filename
        ttf_fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), ttf_filename)
        font_id = QFontDatabase.addApplicationFont(ttf_fname)
        if font_id >= 0:
            font_families = QFontDatabase.applicationFontFamilies(font_id)
        else:
            cache = self.loaded_fonts.get(cache_key, None)
            if cache is None:
                raise OSError("Could not load ttf file for icon font.")
            self.char_map = cache["char_map"]
            self.font_name = cache["font_name"]
            return

        self.font_name = font_families[0]

        filename = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), charmap_filename
        )
        with open(filename, 'r') as codes:
            self.char_map = json.load(codes, object_hook=hook)

        self.loaded_fonts[cache_key] = {
            "char_map": self.char_map,
            "font_name": self.font_name,
        }

    def get_char_for_name(self, name: str) -> str:
        """
        Get a character icon for the given name from the character map.

        Parameters
        ----------
        name : str
            The user-friendly icon name.

        Returns
        -------
        str
            The Qt-facing icon text to use with the font.
        """
        if name in self.char_map:
            return self.char_map[name]
        raise ValueError("Invalid icon name for font.")

    def _load_font_if_needed(self) -> bool:
        """
        Load the configured font if a QApplication is available
        and the font was not already loaded.

        Returns
        -------
        bool
            Readiness indicator - return True if the font has been loaded.
        """
        if QtWidgets.QApplication.instance() is None:
            return False

        if not self.char_map:
            self.load_font(self.font_file, self.charmap_file)

        # If it was loaded correctly, the char map will be populated:
        return bool(self.char_map)

    def font(self, size: int) -> Optional[QtGui.QFont]:
        """
        Load the font at a given pixel size.

        Returns
        -------
        QtGui.QFont or None
            The font, if available.  If a QApplication is not yet created,
            None will be returned.
        """
        if not self._load_font_if_needed():
            return None

        font = QFont(self.font_name)
        font.setPixelSize(int(size))
        return font

    def icon(self, name, color=None) -> Optional[QtGui.QIcon]:
        """
        Retrieve the icon given a name and color.

        Parameters
        ----------
        name : str
            The Icon string identifier.
            Icon strings can be found at: https://fontawesome.com/icons?d=gallery

        color : QColor, Optional
            The base color to use when constructing the Icon. Default is QColor(90, 90, 90).

        Returns
        -------
        QIcon
            The desired Icon.  ``None`` if a QApplication is not yet available.
        """
        if not self._load_font_if_needed():
            return None

        char = self.get_char_for_name(name)
        engine = CharIconEngine(self, char, color)
        return QIcon(engine)


class CharIconEngine(QIconEngine):
    """Subclass of QIconEngine that is designed to draw characters from icon fonts."""

    def __init__(self, icon_font, char, color=None):
        super(CharIconEngine, self).__init__()
        self.icon_font = icon_font
        self.char = char
        if color is None:
            self._base_color = QColor(90, 90, 90)
        else:
            self._base_color = color
        self._disabled_color = QColor.fromHslF(self._base_color.hueF(), self._base_color.saturationF(),
                                               max(min(self._base_color.lightnessF() + 0.25, 1.0), 0.0))

    def paint(self, painter, rect, mode, state):
        painter.save()
        if mode == QIcon.Disabled:
            color = self._disabled_color
        else:
            color = self._base_color
        painter.setPen(color)
        scale_factor = 1.0
        draw_size = 0.875 * qRound(rect.height() * scale_factor)
        painter.setFont(self.icon_font.font(draw_size))
        painter.setOpacity(1.0)
        painter.drawText(rect, int(Qt.AlignCenter | Qt.AlignVCenter), self.char)
        painter.restore()

    def pixmap(self, size, mode, state):
        pm = QPixmap(size)
        pm.fill(Qt.transparent)
        self.paint(QPainter(pm), QRect(QPoint(0, 0), size), mode, state)
        return pm



# Unit Tests for the PyDM drawing widgets

import os
from logging import ERROR
import pytest

from qtpy.QtGui import QColor, QBrush, QPixmap
from qtpy.QtWidgets import QApplication
from qtpy.QtCore import Property, Qt, QPoint, QSize
from qtpy.QtDesigner import QDesignerFormWindowInterface

from ...widgets.base import PyDMWidget
from ...widgets.drawing import (deg_to_qt, qt_to_deg, PyDMDrawing,
                                PyDMDrawingLine, PyDMDrawingImage,
                                PyDMDrawingRectangle, PyDMDrawingTriangle,
                                PyDMDrawingEllipse,
                                PyDMDrawingCircle, PyDMDrawingArc,
                                PyDMDrawingPie, PyDMDrawingChord,
                                PyDMDrawingPolygon, PyDMDrawingPolyline,
                                PyDMDrawingIrregularPolygon)

from ...utilities.stylesheet import apply_stylesheet


# --------------------
# POSITIVE TEST CASES
# --------------------

# # -------------
# # PyDMDrawing
# # -------------
@pytest.mark.parametrize("deg, expected_qt_deg", [
    (0, 0),
    (1, 16),
    (-1, -16),
])
def test_deg_to_qt(deg, expected_qt_deg):
    """
    Test the conversion from degrees to Qt degrees.

    Expectations:
    The angle measurement in degrees is converted correctly to Qt degrees, which are 16 times more than the degree
    value, i.e. 1 degree = 16 Qt degrees.

    Parameters
    ----------
    deg : int, float
        The angle value in degrees

    expected_qt_deg : int, floag
        The expected Qt degrees after the conversion
    """
    assert deg_to_qt(deg) == expected_qt_deg


@pytest.mark.parametrize("qt_deg, expected_deg", [
    (0, 0),
    (16, 1),
    (-16, -1),
    (-32.0, -2),
    (16.16, 1.01)
])
def test_qt_to_deg(qt_deg, expected_deg):
    """
       Test the conversion from Qt degrees to degrees.

       Expectations:
       The angle measurement in Qt degrees is converted correctly to degrees, which are 16 times less than the Qt degree
       value, i.e. 1 Qt degree = 1/16 degree

       Parameters
       ----------
       qt_deg : int, float
           The angle value in Qt degrees

       expected_deg : int, floag
           The expected degrees after the conversion
       """
    assert qt_to_deg(qt_deg) == expected_deg


def test_pydmdrawing_construct(qtbot):
    """
    Test the construction of a PyDM base object.

    Expectations:
    Attributes are assigned with the appropriate default values.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    assert pydm_drawing.alarmSensitiveBorder is False
    assert pydm_drawing._rotation == 0.0
    assert pydm_drawing._brush.style() == Qt.SolidPattern
    assert pydm_drawing._painter
    assert pydm_drawing._pen.style() == pydm_drawing._pen_style == Qt.NoPen
    assert pydm_drawing._pen_width == 0
    assert pydm_drawing._pen_color == QColor(0, 0, 0)


def test_pydmdrawing_sizeHint(qtbot):
    """
    Test the default size of the widget.

    Expectations:
    The size hint is a fixed size.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    assert pydm_drawing.sizeHint() == QSize(100, 100)


@pytest.mark.parametrize("alarm_sensitive_content", [
    True,
    False,
])
def test_pydmdrawing_paintEvent(qtbot, signals, alarm_sensitive_content):
    """
    Test the paintEvent handling of the widget. This test method will also execute PyDMDrawing alarm_severity_changed
    and draw_item().

    Expectations:
    The paintEvent will be triggered, and the widget's brush color is correctly set.

    NOTE: This test depends on the default stylesheet having different values for 'qproperty-brush' for different alarm states of PyDMDrawing.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    alarm_sensitive_content : bool
        True if the widget will be redraw with a different color if an alarm is triggered; False otherwise.
    """
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    qtbot.addWidget(main_window)
    pydm_drawing = PyDMDrawing(parent=main_window, init_channel='fake://tst')
    qtbot.addWidget(pydm_drawing)
    pydm_drawing.alarmSensitiveContent = alarm_sensitive_content
    brush_before = pydm_drawing.brush.color().name()
    signals.new_severity_signal.connect(pydm_drawing.alarmSeverityChanged)
    signals.new_severity_signal.emit(PyDMWidget.ALARM_MAJOR)

    brush_after = pydm_drawing.brush.color().name()
    if alarm_sensitive_content:
        assert brush_before != brush_after
    else:
        assert brush_before == brush_after


@pytest.mark.parametrize("widget_width, widget_height, expected_results", [
    (4.0, 4.0, (2.0, 2.0)),
    (1.0, 1.0, (0.5, 0.5)),
    (0, 0, (0, 0))
])
def test_pydmdrawing_get_center(qtbot, monkeypatch, widget_width, widget_height,
                                expected_results):
    """
    Test the calculation of the widget's center from its width and height.

    Expectations:
    The center of the widget is correctly calculated.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override default attribute values
    widget_width : int, float
        The width of the widget
    widget_height : int, float
        The height of the widget
    expected_results : tuple
        The location of the center. This is a tuple of the distance from the width and that from the height.
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: widget_width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: widget_height)

    assert pydm_drawing.get_center() == expected_results


@pytest.mark.parametrize(
    "width, height, rotation_deg, pen_width, has_border, max_size, force_no_pen, expected",
    [
        # Zero rotation, with typical width, height, pen_width, and variable max_size, has_border, and force_no_pen
        # width > height
        (25.53, 10.35, 0.0, 2, True, True, True,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 0.0, 2, True, True, False,
         (-10.765, -3.175, 21.53, 6.35)),
        (25.53, 10.35, 0.0, 2, True, False, True,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 0.0, 2, True, False, False,
         (-10.765, -3.175, 21.53, 6.35)),
        (25.53, 10.35, 0.0, 2, False, True, True,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 0.0, 2, False, True, False,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 0.0, 2, False, False, True,
         (-12.765, -5.175, 25.53, 10.35)),

        # width < height
        (10.35, 25.53, 0.0, 2, True, True, True,
         (-5.175, -12.765, 10.35, 25.53)),
        (10.35, 25.53, 0.0, 2, True, True, False,
         (-3.175, -10.765, 6.35, 21.53)),
        (10.35, 25.53, 0.0, 2, True, False, True,
         (-5.175, -12.765, 10.35, 25.53)),
        (10.35, 25.53, 0.0, 2, True, False, False,
         (-3.175, -10.765, 6.35, 21.53)),
        (10.35, 25.53, 0.0, 2, False, True, True,
         (-5.175, -12.765, 10.35, 25.53)),
        (10.35, 25.53, 0.0, 2, False, True, False,
         (-5.175, -12.765, 10.35, 25.53)),
        (10.35, 25.53, 0.0, 2, False, False, True,
         (-5.175, -12.765, 10.35, 25.53)),

        # width == height
        (
        10.35, 10.35, 0.0, 2, True, True, True, (-5.175, -5.175, 10.35, 10.35)),
        (10.35, 10.35, 0.0, 2, True, True, False, (-3.175, -3.175, 6.35, 6.35)),
        (10.35, 10.35, 0.0, 2, True, False, True,
         (-5.175, -5.175, 10.35, 10.35)),
        (
        10.35, 10.35, 0.0, 2, True, False, False, (-3.175, -3.175, 6.35, 6.35)),
        (10.35, 10.35, 0.0, 2, False, True, True,
         (-5.175, -5.175, 10.35, 10.35)),
        (10.35, 10.35, 0.0, 2, False, True, False,
         (-5.175, -5.175, 10.35, 10.35)),
        (10.35, 10.35, 0.0, 2, False, False, True,
         (-5.175, -5.175, 10.35, 10.35)),

        # Variable rotation, max_size, and force_no_pen, has_border is True
        (25.53, 10.35, 45.0, 2, True, True, True,
         (-5.207, -2.111, 10.415, 4.222)),
        (25.53, 10.35, 145.0, 2, True, True, True,
         (-5.714, -2.316, 11.428, 4.633)),
        (25.53, 10.35, 90.0, 2, True, True, False,
         (-3.175, -0.098, 6.35, 0.196)),
        (25.53, 10.35, 180.0, 2, True, False, True,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 270.0, 2, True, False, False,
         (-10.765, -3.175, 21.53, 6.35)),
        (25.53, 10.35, 360.0, 2, False, True, True,
         (-12.765, -5.175, 25.53, 10.35)),
        (25.53, 10.35, 0.72, 2, False, True, False,
         (-12.382, -5.02, 24.764, 10.04)),
        (25.53, 10.35, 71.333, 2, False, False, True,
         (-12.765, -5.175, 25.53, 10.35)),
    ])
def test_pydmdrawing_get_bounds(qtbot, monkeypatch, width, height, rotation_deg,
                                pen_width, has_border, max_size,
                                force_no_pen, expected):
    """
    Test the useful area calculations and compare the resulted tuple to the expected one.

    Expectations:
    The drawable area boundaries are correctly calculated.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override default attribute values
    max_size : bool
        If True, draw the widget within the maximum rectangular dimensions given by ```get_inner_max```. If False,
        draw the widget within the user-provided width and height
    force_no_pen : bool
        If True, consider the pen width while calculating the bounds. If False, do not take into account the pen width
    expected : tuple
        The (x, y) coordinates of the starting point, and the maximum width and height of the rendered image
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    pydm_drawing._rotation = rotation_deg
    pydm_drawing._pen_width = pen_width
    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    if has_border:
        monkeypatch.setattr(PyDMDrawing, "has_border", lambda *args: True)
    else:
        monkeypatch.setattr(PyDMDrawing, "has_border", lambda *args: False)

    calculated_bounds = pydm_drawing.get_bounds(max_size, force_no_pen)
    calculated_bounds = tuple(
        [round(x, 3) if isinstance(x, float) else x for x in calculated_bounds])
    assert calculated_bounds == expected


@pytest.mark.parametrize("pen_style, pen_width, expected_result", [
    (Qt.NoPen, 0, False),
    (Qt.NoPen, 1, False),
    (Qt.SolidLine, 0, False),
    (Qt.DashLine, 0, False),
    (Qt.SolidLine, 1, True),
    (Qt.DashLine, 10, True)
])
def test_pydmdrawing_has_border(qtbot, pen_style, pen_width, expected_result):
    """
    Test the determination whether the widget will be drawn with a border, taking into account the pen style and width

    Expectations:
    The widget has a border if the pen style is not Qt.NoPen, and the pen width is greater than 0.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    pen_style : PenStyle
        The style (patterns) of the pen
    pen_width : int
        The thickness of the pen's lines
    expected_result : bool
        True if the widget has a border, False otherwise
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    pydm_drawing.penStyle = pen_style
    pydm_drawing.penWidth = pen_width

    assert pydm_drawing.has_border() == expected_result


@pytest.mark.parametrize("width, height, expected_result", [
    (10, 15, False),
    (10.5, 22.333, False),
    (-10.333, -10.332, False),
    (10.333, 10.333, True),
    (-20.777, -20.777, True),
    (70, 70, True),
])
def test_pydmdrawing_is_square(qtbot, monkeypatch, width, height,
                               expected_result):
    """
    Check if the widget has the same width and height values.

    Expectations:
    The widget's squareness checking returns True if its width and height are the same; False otherwise.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override dialog behaviors
    width : int, float
        The width of the widget
    height : int, float
        The height of a widget
    expected_result
        True if the widget has equal width and height; False otherwise
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    assert pydm_drawing.is_square() == expected_result


@pytest.mark.parametrize("width, height, rotation_deg, expected", [
    (25.53, 10.35, 0.0, (25.53, 10.35)),
    (10.35, 25.53, 0.0, (10.35, 25.53)),
    (25.53, 10.35, 45.0, (10.415, 4.222)),
    (10.35, 25.53, 45.0, (4.222, 10.415)),
    (10.35, 25.53, 360.0, (10.35, 25.53)),
    (10.35, 25.53, -45.0, (4.222, 10.415)),
    (10.35, 25.53, -270.0, (4.196, 10.35)),
    (10.35, 25.53, -360.0, (10.35, 25.53)),
])
def test_pydmdrawing_get_inner_max(qtbot, monkeypatch, width, height,
                                   rotation_deg, expected):
    """
    Test the calculation of the inner rectangle in a rotated rectangle.

    Expectations:
    The returned inner rectangle's width and height are in a tuple, and must match with the values expected.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override existing method behaviors
    width : int, float
        The width of the rotated rectangle
    height : int, float
        The height of the rotated rectangle
    rotation_deg : float
        The rectangle's rotation angle (in degrees)
    expected : tuple
        The tuple containing the width and height of the inner rectangle
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    pydm_drawing._rotation = rotation_deg
    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    calculated_inner_max = pydm_drawing.get_inner_max()
    calculated_inner_max = tuple(
        [round(x, 3) if isinstance(x, float) else x for x in
         calculated_inner_max])
    assert calculated_inner_max == expected


def test_pydmdrawing_properties_and_setters(qtbot):
    """
    Test the PyDMDrawing base class properties and setters.

    Expectations:
    Attribute values are to be retained and retrieved correctly.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    assert pydm_drawing.penWidth == 0
    assert pydm_drawing.penColor == QColor(0, 0, 0)
    assert pydm_drawing.rotation == 0.0
    assert pydm_drawing._brush.style() == Qt.SolidPattern
    assert pydm_drawing.penStyle == Qt.NoPen
    assert pydm_drawing.penCapStyle == Qt.SquareCap
    assert pydm_drawing.penJoinStyle == Qt.MiterJoin

    # The pen width will retain the previously set value if a negative value
    # is attempted to be assigned to it
    pydm_drawing.penWidth = -1
    assert pydm_drawing.penWidth == 0

    pydm_drawing.penWidth = 5
    pydm_drawing.penWidth = -1
    assert pydm_drawing.penWidth == 5

    pydm_drawing.penJoinStyle = Qt.RoundJoin
    assert pydm_drawing.penJoinStyle == Qt.RoundJoin

    pydm_drawing.penCapStyle = Qt.RoundCap
    assert pydm_drawing.penCapStyle == Qt.RoundCap

    pydm_drawing.penColor = QColor(255, 0, 0)
    pydm_drawing.rotation = 99.99
    pydm_drawing.brush = QBrush(Qt.Dense3Pattern)

    assert pydm_drawing.penColor == QColor(255, 0, 0)
    assert pydm_drawing.rotation == 99.99
    assert pydm_drawing._brush.style() == Qt.Dense3Pattern


# # ----------------
# # PyDMDrawingLine
# # ----------------
@pytest.mark.parametrize("alarm_sensitive_content", [
    True,
    False,
])
def test_pydmdrawingline_draw_item(qtbot, signals, alarm_sensitive_content):
    """
    Test PyDMDrawingLine base class drawing handling.

    Expectations:
    The focus manipulation of the base widget object triggers the draw_item() method of the PyDMDrawLine object.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        To emit the alarm severity change signal in order to make an appearance change for the widget, thus triggering
        a redraw
    alarm_sensitive_content : bool
        True if the widget will be redraw with a different color if an alarm is triggered; False otherwise
    """
    pydm_drawingline = PyDMDrawingLine(init_channel='fake://tst')
    qtbot.addWidget(pydm_drawingline)

    pydm_drawingline.alarmSensitiveContent = alarm_sensitive_content
    signals.new_severity_signal.connect(pydm_drawingline.alarmSeverityChanged)
    signals.new_severity_signal.emit(PyDMWidget.ALARM_MAJOR)

    with qtbot.waitExposed(pydm_drawingline):
        pydm_drawingline.show()
    qtbot.waitUntil(lambda: pydm_drawingline.isEnabled(), timeout=5000)
    pydm_drawingline.setFocus()

    def wait_focus():
        return pydm_drawingline.hasFocus()

    qtbot.waitUntil(wait_focus, timeout=5000)


# # -----------------
# # PyDMDrawingImage
# # -----------------
def test_pydmdrawingimage_construct(qtbot):
    """
    Test the construct of a PyDMDrawingImage object.

    Expectations:
    The default attribute values are correctly set.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawingimage = PyDMDrawingImage()
    qtbot.addWidget(pydm_drawingimage)

    assert pydm_drawingimage._pixmap is not None
    assert pydm_drawingimage._aspect_ratio_mode == Qt.KeepAspectRatio
    assert pydm_drawingimage.filename == ""

    base_path = os.path.dirname(__file__)
    test_file = os.path.join(base_path, '..', '..', '..', 'examples', 'drawing',
                             'SLAC_logo.jpeg')
    pydm_drawingimage2 = PyDMDrawingImage(filename=test_file)
    qtbot.addWidget(pydm_drawingimage2)

    pydm_drawingimage3 = PyDMDrawingImage(filename=os.path.abspath(test_file))
    qtbot.addWidget(pydm_drawingimage3)

    pydm_drawingimage4 = PyDMDrawingImage(filename="foo")
    qtbot.addWidget(pydm_drawingimage4)

    test_gif = os.path.join(base_path, '..', '..', '..', 'examples', 'drawing',
                             'test.gif')
    pydm_drawingimage5 = PyDMDrawingImage(filename=test_gif)
    pydm_drawingimage5.movie_finished()
    qtbot.addWidget(pydm_drawingimage5)

    pydm_drawingimage5.filename = test_file
    pydm_drawingimage5.movie_frame_changed(-1)
    pydm_drawingimage5.movie_finished()


def test_pydmdrawingimage_get_designer_window(qtbot):
    """
    Test getting the designer window that owns the widget. Currently, only test with the parent window being None.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    parent = None
    pydm_drawingimage = PyDMDrawingImage(parent=parent)
    qtbot.addWidget(pydm_drawingimage)

    designer_window = pydm_drawingimage.get_designer_window()

    if parent is None:
        assert designer_window is None
    elif isinstance(parent, QDesignerFormWindowInterface):
        assert designer_window == parent
    else:
        assert designer_window == parent.parent()


def test_pydmdrawingimage_test_properties_and_setters(qtbot):
    """
    Test the PyDMDrawing base class properties and setters.

    Expectations:
    Attribute values are to be retained and retrieved correctly.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawingimage = PyDMDrawingImage()
    qtbot.addWidget(pydm_drawingimage)

    pydm_drawingimage.aspectRatioMode = Qt.KeepAspectRatioByExpanding
    assert pydm_drawingimage.aspectRatioMode == Qt.KeepAspectRatioByExpanding


@pytest.mark.parametrize("is_pixmap_empty", [
    True,
    False,
])
def test_pydmdrawingimage_size_hint(qtbot, monkeypatch, is_pixmap_empty):
    """
    Test the size hint of a PyDMDrawingImage object.

    Expectations:
    If the image is empty, the widget assumes the default size of width == height == 100 pixels. If not, the widget
    takes the size from the off screen image presentation (QPixmap).

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    is_pixmap_empty : bool
        True if the image presentation is empty; False otherwise
    """
    pydm_drawingimage = PyDMDrawingImage()
    qtbot.addWidget(pydm_drawingimage)

    if is_pixmap_empty:
        monkeypatch.setattr(QSize, "isEmpty", lambda *args: True)
    else:
        monkeypatch.setattr(QPixmap, "size", lambda *args: QSize(125, 125))

    size_hint = pydm_drawingimage.sizeHint()
    assert size_hint == QSize(100,
                              100) if is_pixmap_empty else size_hint == pydm_drawingimage._pixmap.size()


@pytest.mark.parametrize("width, height, pen_width", [
    (7.7, 10.2, 0),
    (10.2, 7.7, 0),
    (5.0, 5.0, 0),
    (10.25, 10.25, 1.5),
    (10.25, 100.0, 5.125),
    (100.0, 10.25, 5.125),
])
def test_pydmdrawingimage_draw_item(qapp, qtbot, monkeypatch, width, height,
                                    pen_width):
    """
    Test the rendering of a PyDMDrawingImage object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width : int
        The width of the pen stroke
    """
    pydm_drawingimage = PyDMDrawingImage()
    qtbot.addWidget(pydm_drawingimage)

    pydm_drawingimage.penWidth = pen_width

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingimage.show()


# # ---------------------
# # PyDMDrawingRectangle
# # ---------------------
@pytest.mark.parametrize("width, height, pen_width", [
    (7.7, 10.2, 0),
    (10.2, 7.7, 0),
    (5.0, 5.0, 0),
    (10.25, 10.25, 1.5),
    (10.25, 100.0, 5.125),
    (100.0, 10.25, 5.125),
])
def test_pydmdrawingrectangle_draw_item(qapp, qtbot, monkeypatch, width, height,
                                        pen_width):
    """
    Test the rendering of a PyDMDrawingRectangle object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width : int
        The width of the pen stroke
    """
    pydm_drawingrectangle = PyDMDrawingRectangle()
    qtbot.addWidget(pydm_drawingrectangle)

    pydm_drawingrectangle.penWidth = pen_width

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingrectangle.show()


# # ---------------------
# # PyDMDrawingTriangle
# # ---------------------
@pytest.mark.parametrize("width, height, pen_width", [
    (7.7, 10.2, 0),
    (10.2, 7.7, 0),
    (5.0, 5.0, 0),
    (10.25, 10.25, 1.5),
    (10.25, 100.0, 5.125),
    (100.0, 10.25, 5.125),
])
def test_pydmdrawingtriangle_draw_item(qapp, qtbot, monkeypatch, width, height,
                                       pen_width):
    """
    Test the rendering of a PyDMDrawingTriangle object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width : int
        The width of the pen stroke
    """
    pydm_drawingtriangle = PyDMDrawingTriangle()
    qtbot.addWidget(pydm_drawingtriangle)

    pydm_drawingtriangle.penWidth = pen_width

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingtriangle.show()

# # -------------------
# # PyDMDrawingEllipse
# # -------------------
@pytest.mark.parametrize("width, height, pen_width", [
    (5.0, 5.0, 0),
    (10.25, 10.25, 1.5),
    (10.25, 100.0, 5.125),
])
def test_pydmdrawingellipse_draw_item(qapp, qtbot, monkeypatch, width, height,
                                      pen_width):
    """
    Test the rendering of a PyDMDrawingEllipse object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width : int
        The width of the pen stroke
    """
    pydm_dymdrawingellipse = PyDMDrawingEllipse()
    qtbot.addWidget(pydm_dymdrawingellipse)

    pydm_dymdrawingellipse.penWidth = pen_width

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_dymdrawingellipse.show()


# # ------------------
# # PyDMDrawingCircle
# # ------------------
@pytest.mark.parametrize("width, height, expected_radius", [
    (5.0, 5.0, 2.5),
    (10.25, 10.25, 5.125),
    (10.25, 100.0, 5.125),
])
def test_pydmdrawingcircle_calculate_radius(qtbot, width, height,
                                            expected_radius):
    """
    Test the calculation of a PyDMDrawingCircle's radius.

    Expectations:
    Given the width and height of the circle, the calculated radius will match with the expected value.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    expected_radius : int, float
        The expected radius calculated from the given width and height
    """
    pydm_dymdrawingcircle = PyDMDrawingCircle()
    qtbot.addWidget(pydm_dymdrawingcircle)

    calculated_radius = pydm_dymdrawingcircle._calculate_radius(width, height)
    assert calculated_radius == expected_radius


@pytest.mark.parametrize("width, height, pen_width", [
    (5.0, 5.0, 0),
    (10.25, 10.25, 1.5),
    (10.25, 100.0, 5.125),
])
def test_pydmdrawingcircle_draw_item(qapp, qtbot, monkeypatch, width, height,
                                     pen_width):
    """
    Test the rendering of a PyDMDrawingCircle object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width : int
        The width of the pen stroke
    """
    pydm_dymdrawingcircle = PyDMDrawingCircle()
    qtbot.addWidget(pydm_dymdrawingcircle)

    pydm_dymdrawingcircle.penWidth = pen_width

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_dymdrawingcircle.show()


# # ---------------
# # PyDMDrawingArc
# # ---------------
def test_pydmdrawingarc_construct(qtbot):
    """
    Test the construct of a PyDMDrawingArc widget.

    Expectations:
    The default attribute values are as expected.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_drawingarc = PyDMDrawingArc()
    qtbot.addWidget(pydm_drawingarc)

    assert pydm_drawingarc._pen_style == Qt.SolidLine
    assert pydm_drawingarc._pen_width == 1.0
    assert pydm_drawingarc._start_angle == 0
    assert pydm_drawingarc._span_angle == deg_to_qt(90)


@pytest.mark.parametrize("width, height, start_angle_deg, span_angle_deg", [
    (10.333, 11.777, 0, 0),
    (10.333, 10.333, 0, 0),
    (10.333, 10.333, 0, 45),
    (10.333, 11.777, 0, 45),
    (10.333, 11.777, 0, -35),
    (10.333, 11.777, 11, 45),
    (10.333, 11.777, -11, -25),
])
def test_pydmdrawingarc_draw_item(qapp, qtbot, monkeypatch, width, height,
                                  start_angle_deg, span_angle_deg):
    """
    Test the rendering of a PyDMDrawingArc object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    start_angle_deg : int
        The start angle in degrees
    span_angle_deg : int
        The span angle in degrees
    """
    pydm_drawingarc = PyDMDrawingArc()
    qtbot.addWidget(pydm_drawingarc)

    pydm_drawingarc.startAngle = start_angle_deg
    pydm_drawingarc.spanAngle = span_angle_deg

    assert pydm_drawingarc.startAngle == start_angle_deg
    assert pydm_drawingarc.spanAngle == span_angle_deg

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingarc.show()


# # ---------------
# # PyDMDrawingPie
# # ---------------
@pytest.mark.parametrize(
    "width, height, pen_width, rotation_deg, start_angle_deg, span_angle_deg", [
        (10.333, 11.777, 0, 0, 0, 0),
        (10.333, 10.333, 0, 0, 0, 0),
        (10.333, 11.777, 0, 0, 0, 45),
        (10.333, 11.777, 0, 0, 0, -35),
        (10.333, 11.777, 3, 15.333, 0, 0),
        (10.333, 11.777, 3, 15.333, 0, 45),
        (10.333, 11.777, 3, 15.333, 0, -35),
        (10.333, 11.777, 3, 15.333, 11, 45),
        (10.333, 11.777, 3, 15.333, -11, -25),
    ])
def test_pydmdrawingpie_draw_item(qapp, qtbot, monkeypatch, width, height, pen_width,
                                  rotation_deg, start_angle_deg,
                                  span_angle_deg):
    """
    Test the rendering of a PyDMDrawingPie object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
       Window for widget testing
    monkeypatch : fixture
       To override attribute values
    width : int, float
       The width to the widget
    height : int, float
       The height of the widget
    pen_width: int
        The thickness of the pen stroke
    rotation_deg : int
        The rotation in degrees
    start_angle_deg : int
       The start angle in degrees
    span_angle_deg : int
       The span angle in degrees
    """
    pydm_drawingpie = PyDMDrawingPie()
    qtbot.addWidget(pydm_drawingpie)

    pydm_drawingpie._pen_width = pen_width
    pydm_drawingpie._rotation = rotation_deg
    pydm_drawingpie._start_angle = start_angle_deg
    pydm_drawingpie._span_angle = span_angle_deg

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingpie.show()


# # -----------------
# # PyDMDrawingChord
# # -----------------
@pytest.mark.parametrize(
    "width, height, pen_width, rotation_deg, start_angle_deg, span_angle_deg", [
        (10.333, 11.777, 0, 0, 0, 0),
        (10.333, 10.333, 0, 0, 0, 0),
        (10.333, 11.777, 0, 0, 0, 45),
        (10.333, 11.777, 0, 0, 0, -35),
        (10.333, 11.777, 3, 15.333, 0, 0),
        (10.333, 11.777, 3, 15.333, 0, 45),
        (10.333, 11.777, 3, 15.333, 0, -35),
        (10.333, 11.777, 3, 15.333, 11, 45),
        (10.333, 11.777, 3, 15.333, -11, -25),
    ])
def test_pydmdrawingchord_draw_item(qapp, qtbot, monkeypatch, width, height,
                                    pen_width, rotation_deg, start_angle_deg,
                                    span_angle_deg):
    """
    Test the rendering of a PyDMDrawingChord object.

    Expectations:
    The drawing of the object takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    pen_width: int
        The thickness of the pen stroke
    rotation_deg : int
        The rotation in degrees
    start_angle_deg : int
        The start angle in degrees
    span_angle_deg : int
        The span angle in degrees
    """
    pydm_drawingchord = PyDMDrawingChord()
    qtbot.addWidget(pydm_drawingchord)

    pydm_drawingchord._pen_width = pen_width
    pydm_drawingchord._rotation = rotation_deg
    pydm_drawingchord._start_angle = start_angle_deg
    pydm_drawingchord._span_angle = span_angle_deg

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawingchord.show()

# # ---------------------
# # PyDMDrawingPolygon
# # ---------------------
@pytest.mark.parametrize("x, y, width, height, num_points, expected_points", [
    (0, 0, 100, 100, 3, [(50.0, 0),(-25, 43.3012),(-25, -43.3012)]),
    (0, 0, 100, 100, 4, [(50.0, 0), (0, 50.0), (-50.0, 0), (0, -50.0)])
])
def test_pydmdrawingpolygon_calculate_drawing_points(qapp, qtbot, x, y, width,
                                                      height, num_points,
                                                      expected_points):
    """
    Test the calculations of the point coordinates of a PyDMDrawingTriangle widget.

    Expectations:
    The calculations match with the expected values.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    x : int, float
        The x-coordinate
    y: int, float
        The y-coordinate
    width : int, float
        The base measurement
    height : int, float
        The height measurement
    num_points : int
        The number of points in the polygon
    expected_points : tuple
        The collection of the x and y coordinate sets
    """
    drawing = PyDMDrawingPolygon()
    qtbot.addWidget(drawing)

    drawing.numberOfPoints = num_points

    assert drawing.numberOfPoints == num_points

    calculated_points = drawing._calculate_drawing_points(x, y,
                                                          width,
                                                          height)

    for idx, p in enumerate(calculated_points):
        assert p.x() == pytest.approx(expected_points[idx][0], 0.1)
        assert p.y() == pytest.approx(expected_points[idx][1], 0.1)

    drawing.show()

# # ---------------------
# # PyDMDrawingPolyline
# # ---------------------
@pytest.mark.parametrize("x, y, width, height, num_points, expected_points", [
    (-1, 27, 389, 3,  2, [(-2, -2),(384, -2)]),
    (301, 230, 99, 20,  3, [(-1, 18),(-1, -1),(97, -1)])
])
def test_pydmdrawingpolyline_getpoints(qapp, qtbot, x, y, width,
                                                      height, num_points,
                                                      expected_points):
    """
    Test the calculations of the point coordinates of a PyDMDrawingPolyline widget.

    Expectations:
    The calculations match with the expected values.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    x : int, float
        The x-coordinate
    y: int, float
        The y-coordinate
    width : int, float
        The base measurement
    height : int, float
        The height measurement
    num_points : int
        The number of points in the polygon
    expected_points : tuple
        The collection of the x and y coordinate sets
    """
    drawing = PyDMDrawingPolyline()
    qtbot.addWidget(drawing)

    drawing.numberOfPoints = num_points

    assert drawing.numberOfPoints == num_points

    the_points = drawing.getPoints()

    for idx, p in enumerate(the_points):
        assert p.x() == pytest.approx(expected_points[idx][0], 0.1)
        assert p.y() == pytest.approx(expected_points[idx][1], 0.1)

    drawing.show()


@pytest.mark.parametrize("width, height, points, num_points", [
    (99, 20, ["-1, 18", "-1, -1", "97, -1"], 3),
    (99, 20, ["-1, 18", "-1, -1", "97, -1", "", " ", "a b"], 3),
    (99, 20, [[-1, 18], (-1, -1), [97, -1]], 3),
    (99, 20, [[-1, 18], (-1, -1)], 2),
    (99, 20, [[-1, 18], (-1, -1.1)], 2),
    (99, 20, [[-1, 18], "-1, -1.1"], 2),
    (99, 20, [[-1, 18], "-1, -1.1", "5"], 2),
    (99, 20, [[-1, 18], "-1, -1.1", ""], 2),
    (99, 20, [[-1, 18], "-1, -1.1", " "], 2),
    (99, 20, [[-1, 18],], 0),
])
def test_pydmdrawingpolyline_setpoints(qapp, qtbot, monkeypatch, width, height, points, num_points):
    """
    Test the rendering of a PyDMDrawingPolyline widget.

    Expectations:
    The drawing of the widget takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    width : int, float
        The width to the widget
    height : int, float
        The height of the widget
    points : [(float, float)]
        Requested vertices of the polygon.
    num_points :int
        The actual number of vertices of the polygon.
    """
    drawing = PyDMDrawingPolyline()
    qtbot.addWidget(drawing)

    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    drawing.setPoints(points)
    assert len(drawing.getPoints()) == num_points

    drawing.show()

@pytest.mark.parametrize("points, num_points", [
    ([[-1, 18], [-1, -1], [97, -1]], 3),
    ([[-1, 18], [-1, -1]], 2)
])
def test_pydmdrawingpolyline_arrows(qapp, qtbot, points, num_points):
    """
    Test the rendering of a PyDMDrawingPolyline widget with arrow options enabled.

    Expectations:
    The drawing of the widget takes place without any problems.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    points : [(float, float)]
        Requested vertices of the polygon.
    num_points :int
        The actual number of vertices of the polygon.
    """
    drawing = PyDMDrawingPolyline()
    qtbot.addWidget(drawing)

    # make sure points seutp correctly before drawing arrows
    drawing.setPoints(points)
    assert len(drawing.getPoints()) == num_points

    # enable all arrow options
    drawing._arrow_end_point_selection = True
    drawing._arrow_start_point_selection = True
    drawing._arrow_mid_point_selection = True
    drawing._arrow_mid_point_flipped = True
    drawing.draw_item(drawing._painter)
    
    drawing.show()

# # ---------------------------
# # PyDMDrawingIrregularPolygon
# # ---------------------------
@pytest.mark.parametrize("num_points, points", [
    (3, [(-2, -2), (384, -2)]),
    (3, [(-2, -2), (384, -2), (-2, -2)]),
    (4, [(-1, 18), (-1, -1), (97, -1)]),
    (4, [(-1, 18), (-1, -1), "97, -1"]),
    (4, [(-1, 18), (-1, -1), "97, -1", "-1, 18"]),
    (4, [(-1, 18), (-1, -1), "97, -1", "-1    18"]),
    (4, [(-1, 18), (-1, -1), "97, -1", "-1    18", (-1, 18)]),
    (4, [(-1, 18), (-1, -1), "97, -1", "5"]),
    (4, [(-1, 18), (-1, -1), "97, -1", ""]),
    (4, [(-1, 18), (-1, -1), "97, -1", " "]),
    (None, [(-2, -2), ]),
])
def test_pydmdrawingirregularpolygon_get_set_resetpoints(qapp, qtbot, num_points, points):
    """
    Test the calculations of the point coordinates of a PyDMDrawingIrregularPolygon widget.

    Expectations:
    The number of points is correct for a closed polygon.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    num_points : int
        The number of points in the polygon
    points : tuple
        The collection of the x and y coordinate sets
    """
    drawing = PyDMDrawingIrregularPolygon()
    qtbot.addWidget(drawing)

    assert drawing.getPoints() == []

    drawing.setPoints(points)
    vertices = drawing.getPoints()
    if num_points is None:
        assert vertices == []
    else:
        assert len(vertices) == num_points, f"points={points}"
        assert vertices[0] == vertices[-1]  # closed polygon

    drawing.resetPoints()
    assert drawing.getPoints() == []

    drawing.show()


# --------------------
# NEGATIVE TEST CASES
# --------------------

# # -------------
# # PyDMDrawing
# # -------------
@pytest.mark.parametrize("width, height, rotation_deg", [
    (0, 10.35, 0.0),
    (10.35, 0, 0.0),
    (0, 0, 45.0),
    (-10.5, 10.35, 15.0),
    (10.35, -5, 17.5),
    (-10.7, -10, 45.50),
])
def test_get_inner_max_neg(qtbot, monkeypatch, caplog, width, height,
                           rotation_deg):
    """
    Test the handling of invalid width and/or height value during the inner rectangle calculations.

    Expectations:
    Invalid values will be logged as errors.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    monkeypatch : fixture
        To override attribute values
    caplog : fixture
        To capture the error logging
    width : int, float
        The width of the widget
    height : int, float
        The height of the widget
    rotation_deg : int, float
        The widget's rotation, in degrees
    """
    pydm_drawing = PyDMDrawing()
    qtbot.addWidget(pydm_drawing)

    pydm_drawing._rotation = rotation_deg
    monkeypatch.setattr(PyDMDrawing, "width", lambda *args: width)
    monkeypatch.setattr(PyDMDrawing, "height", lambda *args: height)

    pydm_drawing.get_inner_max()

    for record in caplog.records:
        assert record.levelno == ERROR

    if width == 0:
        assert "Invalid width. The value must be greater than 0" in caplog.text
    elif height == 0:
        assert "Invalid height. The value must be greater than 0" in caplog.text



# coding: utf-8
# Fixtures for PyDM Unit Tests

import numpy as np
import pytest
import tempfile
import logging

from qtpy.QtCore import QObject, Signal, Slot
from pydm.application import PyDMApplication
from pydm.main_window import PyDMMainWindow
from pydm.data_plugins import PyDMPlugin, add_plugin

logger = logging.getLogger(__name__)
_, file_path = tempfile.mkstemp(suffix=".log")
handler = logging.FileHandler(file_path)
logger.addHandler(handler)


def mock_method(*args, **kwargs):
    pass


PyDMMainWindow.closeEvent = mock_method


class ConnectionSignals(QObject):
    """
    An assortment of signals, to which a unit test can choose from and bind an appropriate slot
    """
    new_value_signal = Signal([float], [int], [str], [np.ndarray])
    connection_state_signal = Signal(bool)
    new_severity_signal = Signal(int)
    write_access_signal = Signal(bool)
    enum_strings_signal = Signal(tuple)
    internal_slider_moved = Signal(int)
    internal_slider_clicked = Signal()
    send_value_signal = Signal([int], [float], [str], [bool], [np.ndarray])
    unit_signal = Signal(str)
    prec_signal = Signal(int)
    upper_ctrl_limit_signal = Signal([float])
    lower_ctrl_limit_signal = Signal([float])
    upper_alarm_limit_signal = Signal([float])
    lower_alarm_limit_signal = Signal([float])
    upper_warning_limit_signal = Signal([float])
    lower_warning_limit_signal = Signal([float])

    def __init__(self):
        super(ConnectionSignals, self).__init__()
        self._value = None
        self._received_values = {}

    def reset(self):
        self._value = None
        self._received_values.clear()

    @property
    def value(self):
        """
        The property to retrieve the value received from a PyDM widget.

        Returns
        -------
        The value received from a PyDM widget.
        """
        return self._value

    @Slot(int)
    @Slot(float)
    @Slot(str)
    @Slot(np.ndarray)
    def receiveValue(self, val):
        """
        The slot to receive the value from a PyDM widget.

        Parameters
        ----------
        val : int, float, str, ndarray
            The value received from a PyDM widget
        """
        self._value = val

    @Slot(object)
    def receive_value(self, name: str, value: object) -> None:
        """
        Slot for receiving a value from a signal.

        Parameters
        ----------
        name : str
            Name to associate with the value received
        value : object
            The value that was sent by the signal
        """
        self._received_values[name] = value

    def __getitem__(self, name):
        return self._received_values[name]


@pytest.fixture(scope="function")
def signals():
    """
    Wraparound signal collection to work as a fixture for all unit tests.
    This fixture has a function scope to ensure we have a fresh fixture, i.e. a new set of signals, for every unit
    test run.

    Returns
    -------
    A collection of signals to bind to slots.
    """
    return ConnectionSignals()


@pytest.fixture(scope='session')
def qapp(qapp_args):
    """
    Fixture for a PyDMApplication app instance.

    Parameters
    ----------
    qapp_args: Arguments for the QApp.

    Returns
    -------
    An instance of PyDMApplication.
    """
    yield PyDMApplication(use_main_window=False, *qapp_args)


@pytest.fixture(scope='session')
def test_plugin():
    # Create test PyDMPlugin with mock protocol
    test_plug = PyDMPlugin
    test_plug.protocol = 'tst'
    add_plugin(test_plug)
    return test_plug



import functools
import weakref
import threading

from numpy import ndarray
from typing import Optional, Callable

from ..utilities.remove_protocol import parsed_address
from qtpy.QtCore import Signal, QObject, Qt
from qtpy.QtWidgets import QApplication
from .. import config
import re

class PyDMConnection(QObject):
    new_value_signal = Signal([float], [int], [str], [bool], [object])
    connection_state_signal = Signal(bool)
    new_severity_signal = Signal(int)
    write_access_signal = Signal(bool)
    enum_strings_signal = Signal(tuple)
    unit_signal = Signal(str)
    prec_signal = Signal(int)
    upper_ctrl_limit_signal = Signal([float], [int])
    lower_ctrl_limit_signal = Signal([float], [int])
    upper_alarm_limit_signal = Signal([float], [int])
    lower_alarm_limit_signal = Signal([float], [int])
    upper_warning_limit_signal = Signal([float], [int])
    lower_warning_limit_signal = Signal([float], [int])
    timestamp_signal = Signal(float)

    def __init__(self, channel, address, protocol=None, parent=None):
        super(PyDMConnection, self).__init__(parent)
        self.protocol = protocol
        self.address = address
        self.connected = False
        self.value = None
        self.listener_count = 0
        self.app = QApplication.instance()

    def add_listener(self, channel):
        self.listener_count = self.listener_count + 1
        if channel.connection_slot is not None:
            self.connection_state_signal.connect(channel.connection_slot, Qt.QueuedConnection)

        if channel.value_slot is not None:
            try:
                self.new_value_signal[int].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[float].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[str].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[bool].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[object].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
        
        if channel.severity_slot is not None:
            self.new_severity_signal.connect(channel.severity_slot, Qt.QueuedConnection)

        if channel.write_access_slot is not None:
            self.write_access_signal.connect(channel.write_access_slot, Qt.QueuedConnection)

        if channel.enum_strings_slot is not None:
            self.enum_strings_signal.connect(channel.enum_strings_slot, Qt.QueuedConnection)

        if channel.unit_slot is not None:
            self.unit_signal.connect(channel.unit_slot, Qt.QueuedConnection)

        if channel.upper_ctrl_limit_slot is not None:
            self.upper_ctrl_limit_signal.connect(channel.upper_ctrl_limit_slot, Qt.QueuedConnection)

        if channel.lower_ctrl_limit_slot is not None:
            self.lower_ctrl_limit_signal.connect(channel.lower_ctrl_limit_slot, Qt.QueuedConnection)

        if channel.upper_alarm_limit_slot is not None:
            self.upper_alarm_limit_signal.connect(channel.upper_alarm_limit_slot, Qt.QueuedConnection)

        if channel.lower_alarm_limit_slot is not None:
            self.lower_alarm_limit_signal.connect(channel.lower_alarm_limit_slot, Qt.QueuedConnection)

        if channel.upper_warning_limit_slot is not None:
            self.upper_warning_limit_signal.connect(channel.upper_warning_limit_slot, Qt.QueuedConnection)

        if channel.lower_warning_limit_slot is not None:
            self.lower_warning_limit_signal.connect(channel.lower_warning_limit_slot, Qt.QueuedConnection)

        if channel.prec_slot is not None:
            self.prec_signal.connect(channel.prec_slot, Qt.QueuedConnection)

        if channel.timestamp_slot is not None:
            self.timestamp_signal.connect(channel.timestamp_slot, Qt.QueuedConnection)

    def remove_listener(self, channel, destroying: Optional[bool] = False) -> None:
        """
        Removes a listener from this PyDMConnection. If there are no more listeners remaining after
        removal, then the PyDMConnection will be closed.

        Parameters
        ----------
        channel: PyDMChannel
            The PyDMChannel containing the signals/slots that were being used to listen to the connected address.
        destroying: bool, optional
            Should be set to True if this method is being invoked from a flow in which the PyDMWidget using this
            channel is being destroyed. Since Qt will automatically handle the disconnect of signals/slots when a
            QObject is destroyed, setting this to True ensures we do not try to do the disconnection a second time.
            If set to False, any active signals/slots on the channel will be manually disconnected here.
        """
        if self._should_disconnect(channel.connection_slot, destroying):
            try:
                self.connection_state_signal.disconnect(channel.connection_slot)
            except TypeError:
                pass

        if self._should_disconnect(channel.value_slot, destroying):
            try:
                self.new_value_signal[int].disconnect(channel.value_slot)
            except TypeError:
                pass
            try:
                self.new_value_signal[float].disconnect(channel.value_slot)
            except TypeError:
                pass
            try:
                self.new_value_signal[str].disconnect(channel.value_slot)
            except TypeError:
                pass
            try:
                self.new_value_signal[bool].disconnect(channel.value_slot)
            except TypeError:
                pass
            try:
                self.new_value_signal[object].disconnect(channel.value_slot)
            except TypeError:
                pass
        
        if self._should_disconnect(channel.severity_slot, destroying):
            try:
                self.new_severity_signal.disconnect(channel.severity_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.write_access_slot, destroying):
            try:
                self.write_access_signal.disconnect(channel.write_access_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.enum_strings_slot, destroying):
            try:
                self.enum_strings_signal.disconnect(channel.enum_strings_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.unit_slot, destroying):
            try:
                self.unit_signal.disconnect(channel.unit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.upper_ctrl_limit_slot, destroying):
            try:
                self.upper_ctrl_limit_signal.disconnect(channel.upper_ctrl_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.lower_ctrl_limit_slot, destroying):
            try:
                self.lower_ctrl_limit_signal.disconnect(channel.lower_ctrl_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.upper_alarm_limit_slot, destroying):
            try:
                self.upper_alarm_limit_signal.disconnect(channel.upper_alarm_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.lower_alarm_limit_slot, destroying):
            try:
                self.lower_alarm_limit_signal.disconnect(channel.lower_alarm_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.upper_warning_limit_slot, destroying):
            try:
                self.upper_warning_limit_signal.disconnect(channel.upper_warning_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.lower_warning_limit_slot, destroying):
            try:
                self.lower_warning_limit_signal.disconnect(channel.lower_warning_limit_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.prec_slot, destroying):
            try:
                self.prec_signal.disconnect(channel.prec_slot)
            except (KeyError, TypeError):
                pass

        if self._should_disconnect(channel.timestamp_slot, destroying):
            try:
                self.timestamp_signal.disconnect(channel.timestamp_slot)
            except (KeyError, TypeError):
                pass

        self.listener_count = self.listener_count - 1
        if self.listener_count < 1:
            self.close()

    @staticmethod
    def _should_disconnect(slot: Callable, destroying: bool):
        """ Return True if the signal/slot should be disconnected, False otherwise """
        if slot is None:
            # Nothing to do if the slot does not exist
            return False
        if not destroying:
            # If the PyDMWidget associated with this slot is not being destroyed, then we do need to
            # manually disconnect the signal/slot
            return True
        if isinstance(slot, functools.partial):
            # If the slot was created as a partial, we also need to manually disconnect it even if the PyDMWidget
            # is being destroyed since Qt does not handle automatic disconnection when a partial is used
            return True
        # This means we are destroying the PyDMWidget and the slot is not a partial, so let Qt
        # handle the disconnect for us
        return False

    def close(self):
        pass


class PyDMPlugin(object):
    protocol = None
    connection_class = PyDMConnection
    designer_online_by_default = False

    def __init__(self):
        self.connections = {}
        self.channels = weakref.WeakSet()
        self.lock = threading.Lock()

    @staticmethod
    def get_parsed_address(channel):
        parsed_addr = parsed_address(channel.address)
        return parsed_addr

    @staticmethod
    def get_full_address(channel):
        parsed_addr = parsed_address(channel.address)

        if parsed_addr:
            full_addr = parsed_addr.netloc + parsed_addr.path
        else: 
            full_addr = None

        return full_addr

    @staticmethod
    def get_address(channel):
        parsed_addr = parsed_address(channel.address)
        addr = parsed_addr.netloc
        
        return addr
    
    @staticmethod
    def get_subfield(channel):
        parsed_addr = parsed_address(channel.address)

        if parsed_addr:
            subfield = parsed_addr.path

            if subfield != '':
                subfield = subfield[1:].split('/')
        else:
            subfield = None
        
        return subfield

    @staticmethod
    def get_connection_id(channel):
        return PyDMPlugin.get_full_address(channel)

    def add_connection(self, channel):
        from pydm.utilities import is_qt_designer
        with self.lock:
            connection_id = self.get_connection_id(channel)
            address = self.get_address(channel)

            # If this channel is already connected to this plugin lets ignore
            if channel in self.channels:
                return

            if (is_qt_designer() and not config.DESIGNER_ONLINE and
                    not self.designer_online_by_default):
                return

            self.channels.add(channel)
            if connection_id in self.connections:
                self.connections[connection_id].add_listener(channel)
            else:
                self.connections[connection_id] = self.connection_class(
                    channel, address, self.protocol)

    def remove_connection(self, channel, destroying=False):
        with self.lock:
            connection_id = self.get_connection_id(channel)
            if connection_id in self.connections and channel in self.channels:
                self.connections[connection_id].remove_listener(
                    channel, destroying=destroying)
                self.channels.remove(channel)
                if self.connections[connection_id].listener_count < 1:
                    self.connections[connection_id].deleteLater()
                    del self.connections[connection_id]




import os
import pytest
import logging

from ... import config
from ...utilities import stylesheet
from qtpy.QtWidgets import QApplication

# The path to the stylesheet used in these unit tests
test_stylesheet_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "..", "test_data", "global_stylesheet.css")


@pytest.fixture(scope='function')
def save_and_restore_pydm_stylesheet():
    """
    A fixture for ensuring that modifications to the PYDM_STYLESHEET environment variable are restored
    even if the test fails. This will prevent any impact to the user's environment, as well as future tests in the run.
    """
    # Back up the stylesheet related variables so they can be restored after the test
    env_backup = os.getenv("PYDM_STYLESHEET", None)
    backup_global_style_path = stylesheet.GLOBAL_STYLESHEET
    os.environ["PYDM_STYLESHEET"] = ""

    yield

    # Restore the user's original stylesheet
    if env_backup:
        os.environ["PYDM_STYLESHEET"] = env_backup
    stylesheet.GLOBAL_STYLESHEET = backup_global_style_path


def test_stylesheet_apply(qtbot, save_and_restore_pydm_stylesheet):
    assert os.getenv("PYDM_STYLESHEET", None) == ""

    # Retrieve instance of the application so we can test with it
    app = QApplication.instance()
    # Ensure that the app stylesheet is clean
    assert not app.styleSheet()

    stylesheet.clear_cache()

    # Exercise apply stylesheet to app
    stylesheet.apply_stylesheet()

    assert app.styleSheet() is not None

    stylesheet.clear_cache()
    # Exercise when there is no stylesheet available
    stylesheet.GLOBAL_STYLESHEET = "invalid_file.none"
    app.setStyleSheet("")
    assert not app.styleSheet()

    # Test apply stylesheet to app
    stylesheet.apply_stylesheet()

    assert not app.styleSheet()


def test_stylesheet_get_style_data(caplog, save_and_restore_pydm_stylesheet):
    assert os.getenv("PYDM_STYLESHEET", None) == ""

    with caplog.at_level(logging.DEBUG):
        # Cleanup the cache
        stylesheet.clear_cache()

        # Check to ensure that the cache is empty
        assert stylesheet.__style_data is None

        # First test the case in which we don't specify the file
        ret = stylesheet._get_style_data()

        assert ret
        assert stylesheet.__style_data == ret

        # Make sure logging capture the error, and have the correct error
        # message
        assert "Opening the default stylesheet" in caplog.text

        caplog.clear()

        # Exercise the cache
        ret = stylesheet._get_style_data()

        assert ret
        assert stylesheet.__style_data == ret

        assert len(caplog.records) == 0

        caplog.clear()

        # Exercise the cache
        ret = stylesheet._get_style_data()

        assert ret
        assert stylesheet.__style_data == ret

        assert len(caplog.records) == 0

        # Clear the cache
        stylesheet.clear_cache()

        # Exercise the invalid file with fallback
        ret = stylesheet._get_style_data("foo.bar.none")

        # Make sure logging capture the error, and have the correct error message
        assert len(caplog.records) == 2
        assert "Error reading the stylesheet file" in caplog.text
        assert "Opening the default stylesheet" in caplog.text

        caplog.clear()
        stylesheet.clear_cache()

        # Exercise the proper loading
        ret = stylesheet._get_style_data(test_stylesheet_path)
        # Make sure logging capture the error, and have the correct error message
        assert "Opening style file" in caplog.text

        caplog.clear()
        stylesheet.clear_cache()

        # Exercise invalid default file
        stylesheet.GLOBAL_STYLESHEET = "invalid_file.none"
        ret = stylesheet._get_style_data()

        # Make sure logging capture the error, and have the correct error message
        if not config.STYLESHEET:
            assert len(caplog.records) == 1
        else:
            assert len(caplog.records) == 2  # Extra message about existing stylesheet
        assert "Cannot find the default stylesheet" in caplog.text



import io
import logging
import numpy as np
from functools import reduce
from p4p.wrapper import Value

logger = logging.getLogger(__name__)


ScalarType = (bool, np.int8, np.int16, np.int32, np.compat.long, np.uint8,
              np.uint16, np.uint32, np.uint64, float, np.double)


codecs = {}


def decompress(structure: Value):
    """
    Performs decompression on the input value if the codec field has been set and is valid. If the data is being sent
    uncompressed already, will just reshape the data using the sizes specified in the dimension field.

    Parameters
    ----------
    structure : Value
        The value we have received from P4P
    """
    if structure is None:
        return

    data = structure.get('value')
    if data is None:
        return
    shape = []
    for dim in structure.get('dimension', []):
        shape.append(dim.get('size'))
    codec = structure.get('codec', {})
    data_type = codec.get('parameters')
    if data_type is None:
        # Assuming data type from data
        dtype = data.dtype
    else:
        dtype = ScalarType[data_type]

    codec_name = codec.get('name')
    uncompressed_size = structure.get('uncompressedSize')

    if not codec_name:
        return none_decompress(data, shape, dtype)

    try:
        return codecs[codec_name](data, shape, dtype, uncompressed_size)
    except Exception:
        logging.exception('Could not run codec decompress for %s', codec_name)
        return data


def none_decompress(data, shape, dtype):
    """ Perform a reshape based on the dimensions that were set """
    return np.frombuffer(data, dtype=dtype).reshape(shape)


def jpeg_decompress(data, shape, dtype, uncompressed_size):
    """ Decompress using Pillow's jpeg decompression """
    return np.array(Image.open(io.BytesIO(data.tobytes())))


def blosc_decompress(data, shape, dtype, uncompressed_size):
    """ Decompress using blosc """
    dec_data = blosc.decompress(data)
    return np.frombuffer(dec_data, dtype=dtype).reshape(shape)


def lz4_decompress(data, shape, dtype, uncompressed_size):
    """ Decompress using lz4 """
    dec_data = block.decompress(data, uncompressed_size)
    return np.frombuffer(dec_data, dtype=dtype).reshape(shape)


def bslz4_decompress(data, shape, dtype, uncompressed_size):
    """ Decompress using bslz4 """
    nelems = reduce((lambda x, y: x * y), shape)
    dec_data = bitshuffle.decompress_lz4(data, (nelems,), np.dtype(dtype))
    return dec_data.reshape(shape)


try:
    import blosc
    codecs['blosc'] = blosc_decompress
except ImportError:
    logger.debug('Blosc codec not available for PVAccess data decompression')

try:
    from lz4 import block
    codecs['lz4'] = lz4_decompress
except ImportError:
    logger.debug('LZ4 codec not available for PVAccess data decompression')

try:
    import bitshuffle
    codecs['bslz4'] = bslz4_decompress
except ImportError:
    logger.debug('BSLZ4 codec not available for PVAccess data decompression')

try:
    from PIL import Image
    codecs['jpeg'] = jpeg_decompress
except ImportError:
    logger.debug('JPEG codec not available for PVAccess data decompression')



from ...utilities.remove_protocol import remove_protocol
from ...utilities.remove_protocol import protocol_and_address
from ...utilities.remove_protocol import parsed_address

def test_remove_protocol():
    out = remove_protocol('foo://bar')
    assert (out == 'bar')

    out = remove_protocol('bar')
    assert (out == 'bar')

    out = remove_protocol('foo://bar://foo2')
    assert (out == 'bar://foo2')

def test_protocol_and_address():
    out = protocol_and_address('foo://bar')
    assert (out == ('foo', 'bar'))

    out = protocol_and_address('foo:/bar')
    assert (out == (None, 'foo:/bar'))

def test_parsed_address():
    out = parsed_address(1)
    assert (out == None)

    out = parsed_address('foo:/bar')
    assert (out == None)

    out = parsed_address('foo://bar')
    assert (out == ('foo', 'bar', '', '', '', ''))





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pydm.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from qtpy import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(677, 465)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 677, 20))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        self.menuHistory = QtWidgets.QMenu(self.menubar)
        self.menuHistory.setObjectName("menuHistory")
        self.menuTools = QtWidgets.QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        self.menuCustomActions = QtWidgets.QMenu(self.menubar)
        self.menuCustomActions.setObjectName("menuCustomActions")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.navbar = QtWidgets.QToolBar(MainWindow)
        self.navbar.setMovable(False)
        self.navbar.setFloatable(False)
        self.navbar.setObjectName("navbar")
        MainWindow.addToolBar(QtCore.Qt.TopToolBarArea, self.navbar)
        self.actionEdit_in_Designer = QtWidgets.QAction(MainWindow)
        self.actionEdit_in_Designer.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionEdit_in_Designer.setObjectName("actionEdit_in_Designer")
        self.actionAbout_PyDM = QtWidgets.QAction(MainWindow)
        self.actionAbout_PyDM.setEnabled(True)
        self.actionAbout_PyDM.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionAbout_PyDM.setObjectName("actionAbout_PyDM")
        self.actionReload_Display = QtWidgets.QAction(MainWindow)
        self.actionReload_Display.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionReload_Display.setObjectName("actionReload_Display")
        self.actionIncrease_Font_Size = QtWidgets.QAction(MainWindow)
        self.actionIncrease_Font_Size.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionIncrease_Font_Size.setObjectName("actionIncrease_Font_Size")
        self.actionDecrease_Font_Size = QtWidgets.QAction(MainWindow)
        self.actionDecrease_Font_Size.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionDecrease_Font_Size.setObjectName("actionDecrease_Font_Size")
        self.actionShow_File_Path_in_Title_Bar = QtWidgets.QAction(MainWindow)
        self.actionShow_File_Path_in_Title_Bar.setCheckable(True)
        self.actionShow_File_Path_in_Title_Bar.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_File_Path_in_Title_Bar.setObjectName("actionShow_File_Path_in_Title_Bar")
        self.actionBack = QtWidgets.QAction(MainWindow)
        self.actionBack.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionBack.setObjectName("actionBack")
        self.actionForward = QtWidgets.QAction(MainWindow)
        self.actionForward.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionForward.setObjectName("actionForward")
        self.actionHome = QtWidgets.QAction(MainWindow)
        self.actionHome.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionHome.setObjectName("actionHome")
        self.actionShow_Navigation_Bar = QtWidgets.QAction(MainWindow)
        self.actionShow_Navigation_Bar.setCheckable(True)
        self.actionShow_Navigation_Bar.setChecked(True)
        self.actionShow_Navigation_Bar.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_Navigation_Bar.setObjectName("actionShow_Navigation_Bar")
        self.actionOpen_File = QtWidgets.QAction(MainWindow)
        self.actionOpen_File.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionOpen_File.setObjectName("actionOpen_File")
        self.actionShow_Menu_Bar = QtWidgets.QAction(MainWindow)
        self.actionShow_Menu_Bar.setCheckable(True)
        self.actionShow_Menu_Bar.setChecked(True)
        self.actionShow_Menu_Bar.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_Menu_Bar.setObjectName("actionShow_Menu_Bar")
        self.actionShow_Status_Bar = QtWidgets.QAction(MainWindow)
        self.actionShow_Status_Bar.setCheckable(True)
        self.actionShow_Status_Bar.setChecked(True)
        self.actionShow_Status_Bar.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_Status_Bar.setObjectName("actionShow_Status_Bar")
        self.actionShow_Connections = QtWidgets.QAction(MainWindow)
        self.actionShow_Connections.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_Connections.setObjectName("actionShow_Connections")
        self.actionShow_Help = QtWidgets.QAction(MainWindow)
        self.actionShow_Help.setShortcutContext(QtCore.Qt.ApplicationShortcut)
        self.actionShow_Help.setObjectName("actionShow_Help")
        self.actionLoadTool = QtWidgets.QAction(MainWindow)
        self.actionLoadTool.setObjectName("actionLoadTool")
        self.actionEnter_Fullscreen = QtWidgets.QAction(MainWindow)
        self.actionEnter_Fullscreen.setObjectName("actionEnter_Fullscreen")
        self.actionDefault_Font_Size = QtWidgets.QAction(MainWindow)
        self.actionDefault_Font_Size.setObjectName("actionDefault_Font_Size")
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setEnabled(True)
        self.actionQuit.setObjectName("actionQuit")
        self.actionSave = QtWidgets.QAction(MainWindow)
        self.actionSave.setObjectName('actionSave')
        self.actionSave.setVisible(False)
        self.actionSave_As = QtWidgets.QAction(MainWindow)
        self.actionSave_As.setObjectName('actionSave_As')
        self.actionSave_As.setVisible(False)
        self.actionLoad = QtWidgets.QAction(MainWindow)
        self.actionLoad.setObjectName('actionLoad')
        self.actionLoad.setVisible(False)
        self.actionChange_Stylesheet = QtWidgets.QAction(MainWindow)
        self.actionChange_Stylesheet.setObjectName("actionChange_Stylesheet")
        self.menuFile.addAction(self.actionOpen_File)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionChange_Stylesheet)
        self.menuFile.addAction(self.actionEdit_in_Designer)
        self.menuFile.addAction(self.actionReload_Display)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionAbout_PyDM)
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionSave_As)
        self.menuFile.addAction(self.actionLoad)
        self.menuFile.addAction(self.actionQuit)
        self.menuView.addAction(self.actionEnter_Fullscreen)
        self.menuView.addAction(self.actionIncrease_Font_Size)
        self.menuView.addAction(self.actionDecrease_Font_Size)
        self.menuView.addAction(self.actionDefault_Font_Size)
        self.menuView.addSeparator()
        self.menuView.addAction(self.actionShow_File_Path_in_Title_Bar)
        self.menuView.addAction(self.actionShow_Navigation_Bar)
        self.menuView.addAction(self.actionShow_Menu_Bar)
        self.menuView.addAction(self.actionShow_Status_Bar)
        self.menuView.addAction(self.actionShow_Connections)
        self.menuView.addAction(self.actionShow_Help)
        self.menuHistory.addAction(self.actionBack)
        self.menuHistory.addAction(self.actionForward)
        self.menuHistory.addAction(self.actionHome)
        self.menuTools.addAction(self.actionLoadTool)
        self.menuTools.addSeparator()
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuHistory.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuCustomActions.menuAction())
        self.navbar.addAction(self.actionBack)
        self.navbar.addAction(self.actionForward)
        self.navbar.addSeparator()
        self.navbar.addAction(self.actionHome)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "PyDM Main Window"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuView.setTitle(_translate("MainWindow", "View"))
        self.menuHistory.setTitle(_translate("MainWindow", "History"))
        self.menuTools.setTitle(_translate("MainWindow", "Tools"))
        self.menuCustomActions.setTitle(_translate("MainWindow", "Actions"))
        self.navbar.setWindowTitle(_translate("MainWindow", "toolBar"))
        self.actionEdit_in_Designer.setText(_translate("MainWindow", "Edit in Designer"))
        self.actionAbout_PyDM.setText(_translate("MainWindow", "About PyDM"))
        self.actionReload_Display.setText(_translate("MainWindow", "Reload Display"))
        self.actionReload_Display.setShortcut(_translate("MainWindow", "Ctrl+R"))
        self.actionIncrease_Font_Size.setText(_translate("MainWindow", "Increase Font Size"))
        self.actionIncrease_Font_Size.setShortcut(_translate("MainWindow", "Ctrl+="))
        self.actionDecrease_Font_Size.setText(_translate("MainWindow", "Decrease Font Size"))
        self.actionDecrease_Font_Size.setShortcut(_translate("MainWindow", "Ctrl+-"))
        self.actionShow_File_Path_in_Title_Bar.setText(_translate("MainWindow", "Show File Path in Title Bar"))
        self.actionBack.setText(_translate("MainWindow", "Back"))
        self.actionBack.setShortcut(_translate("MainWindow", "Ctrl+Left"))
        self.actionForward.setText(_translate("MainWindow", "Forward"))
        self.actionForward.setShortcut(_translate("MainWindow", "Ctrl+Right"))
        self.actionHome.setText(_translate("MainWindow", "Home"))
        self.actionHome.setShortcut(_translate("MainWindow", "Ctrl+H"))
        self.actionShow_Navigation_Bar.setText(_translate("MainWindow", "Show Navigation Bar"))
        self.actionOpen_File.setText(_translate("MainWindow", "Open File..."))
        self.actionOpen_File.setShortcut(_translate("MainWindow", "Ctrl+O"))
        self.actionShow_Menu_Bar.setText(_translate("MainWindow", "Show Menu Bar"))
        self.actionShow_Menu_Bar.setShortcut(_translate("MainWindow", "Ctrl+M"))
        self.actionShow_Status_Bar.setText(_translate("MainWindow", "Show Status Bar"))
        self.actionShow_Connections.setText(_translate("MainWindow", "Show Connections..."))
        self.actionShow_Help.setText(_translate("MainWindow", "View Help for this Display"))
        self.actionLoadTool.setText(_translate("MainWindow", "Load..."))
        self.actionEnter_Fullscreen.setText(_translate("MainWindow", "Enter Fullscreen"))
        self.actionEnter_Fullscreen.setShortcut(_translate("MainWindow", "F11"))
        self.actionDefault_Font_Size.setText(_translate("MainWindow", "Default Font Size"))
        self.actionDefault_Font_Size.setShortcut(_translate("MainWindow", "Ctrl+0"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))
        self.actionSave.setText(_translate("MainWindow", "Save"))
        self.actionSave_As.setText(_translate("MainWindow", "Save As..."))
        self.actionLoad.setText(_translate("MainWindow", "Load"))
        self.actionChange_Stylesheet.setText(_translate("MainWindow", "Change Stylesheet"))



"""
Main Application Module

Contains our PyDMApplication class with core connection and loading logic and
our PyDMMainWindow class with navigation logic.
"""
import os
import signal
import subprocess
import json
import logging
import warnings

from qtpy.QtCore import Qt, QTimer, Slot
from qtpy.QtWidgets import QApplication, QWidget
from .main_window import PyDMMainWindow

from .utilities import which, path_info
from .utilities.stylesheet import apply_stylesheet
from .utilities import connection
from . import config, data_plugins

logger = logging.getLogger(__name__)


class PyDMApplication(QApplication):
    """
    PyDMApplication handles loading PyDM display files, opening
    new windows, and most importantly, establishing and managing
    connections to channels via data plugins.

    Parameters
    ----------
    ui_file : str, optional
        The file path to a PyDM display file (.ui, .py, or .adl).
    command_line_args : list, optional
        A list of strings representing arguments supplied at the command
        line.  All arguments in this list are handled by QApplication,
        in addition to PyDMApplication.
    display_args : list, optional
        A list of command line arguments that should be forwarded to the
        Display class.  This is only useful if a Related Display Button
        is opening up a .py file with extra arguments specified, and
        probably isn't something you will ever need to use when writing
        code that instantiates PyDMApplication.
    perfmon : bool, optional
        Whether or not to enable performance monitoring using 'psutil'.
        When enabled, CPU load information on a per-thread basis is
        periodically printed to the terminal.
    hide_nav_bar : bool, optional
        Whether or not to display the navigation bar (forward/back/home buttons)
        when the main window is first displayed.
    hide_menu_bar: bool, optional
        Whether or not to display the menu bar (File, View)
        when the main window is first displayed.
    hide_status_bar: bool, optional
        Whether or not to display the status bar (general messages and errors)
        when the main window is first displayed.
    read_only: bool, optional
        Whether or not to launch PyDM in a read-only state.
    macros : dict, optional
        A dictionary of macro variables to be forwarded to the display class
        being loaded.
    use_main_window : bool, optional
        If ui_file is note given, this parameter controls whether or not to
        create a PyDMMainWindow in the initialization (Default is True).
    fullscreen : bool, optional
        Whether or not to launch PyDM in a full screen mode.
    homefile : str, optional
        The path to a PyDM file to return to whenever the home button is clicked in the navigation bar.
    """
    # Instantiate our plugins.
    plugins = data_plugins.plugin_modules

    def __init__(self, ui_file=None, command_line_args=[], display_args=[],
                 perfmon=False, hide_nav_bar=False, hide_menu_bar=False,
                 hide_status_bar=False, read_only=False, macros=None,
                 use_main_window=True, stylesheet_path=None, fullscreen=False, home_file=None):
        super(PyDMApplication, self).__init__(command_line_args)
        # Enable High DPI display, if available.
        if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
            self.setAttribute(Qt.AA_UseHighDpiPixmaps)

        data_plugins.set_read_only(read_only)
        self.main_window = None
        self.display_args = display_args
        self.hide_nav_bar = hide_nav_bar
        self.hide_menu_bar = hide_menu_bar
        self.hide_status_bar = hide_status_bar
        self.fullscreen = fullscreen
        self.stylesheet_path = stylesheet_path
        self.perfmon = perfmon

        # The home_file param is set by command line option. If the option wasn't set, try the environment variable
        # from config. Note that this may not be set either, in which case home_file will eventually be set to
        # the first display opened.
        self.home_file = home_file
        if self.home_file is None:
            self.home_file = config.HOME_FILE

        # Open a window if required.
        if ui_file is not None:
            self.make_main_window(stylesheet_path=stylesheet_path, home_file=self.home_file,
                                  macros=macros, command_line_args=command_line_args)
            self.main_window.open(ui_file, macros, command_line_args)
        elif use_main_window:
            self.make_main_window(stylesheet_path=stylesheet_path, home_file=self.home_file,
                                  macros=macros, command_line_args=command_line_args)

        self.had_file = ui_file is not None
        # Re-enable sigint (usually blocked by pyqt)
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        # Performance monitoring
        if perfmon:
            import psutil
            self.perf = psutil.Process()
            self.perf_timer = QTimer()
            self.perf_timer.setInterval(2000)
            self.perf_timer.timeout.connect(self.get_CPU_usage)
            self.perf_timer.start()

    def get_string_encoding(self):
        return os.getenv("PYDM_STRING_ENCODING", "utf_8")

    def exec_(self):
        """
        Execute the QApplication.
        """
        return super(PyDMApplication, self).exec_()

    def is_read_only(self):
        warnings.warn("'PyDMApplication.is_read_only' is deprecated, "
                      "use 'pydm.data_plugins.is_read_only' instead.")
        return data_plugins.is_read_only()

    @Slot()
    def get_CPU_usage(self):
        """
        Prints total CPU usage (in percent), as well as per-thread usage, to the terminal.
        """
        with self.perf.oneshot():
            total_percent = self.perf.cpu_percent(interval=None)
            total_time = sum(self.perf.cpu_times())
            usage = [total_percent * ((t.system_time + t.user_time) / total_time) for t in self.perf.threads()]
        print("Total: {tot}, Per Thread: {percpu}".format(tot=total_percent, percpu=usage))

    def new_pydm_process(self, ui_file, macros=None, command_line_args=None):
        """
        Spawn a new PyDM process and open the supplied file.  Commands to open
        new windows in PyDM typically actually spawn an entirely new PyDM process.
        This keeps each window isolated, so that one window cannot slow
        down or crash another.

        Parameters
        ----------
        ui_file : str
            The path to a .ui, .py, or .adl file to open in the new process.
        macros : dict, optional
            A dictionary of macro variables to supply to the display file
            to be opened.
        command_line_args : list, optional
            A list of command line arguments to pass to the new process.
            Typically, this argument is used by related display buttons
            to pass in extra arguments.  It is probably rare that code you
            write needs to use this argument.
        """
        # Expand user (~ or ~user) and environment variables.
        ui_file = os.path.expanduser(os.path.expandvars(ui_file))
        base_dir, fname, args = path_info(str(ui_file))
        filepath = os.path.join(base_dir, fname)
        filepath_args = args
        pydm_display_app_path = which("pydm")

        if pydm_display_app_path is None:
            if os.environ.get("PYDM_PATH") is not None:
                pydm_display_app_path = os.path.join(os.environ["PYDM_PATH"], "pydm")
            else:
                # Not in the PATH and no ENV VAR pointing to it...
                # Let's try the script folder...
                pydm_display_app_path = os.path.join(os.path.split(os.path.realpath(__file__))[0], "..", "pydm_launcher", "main.py")

        args = [pydm_display_app_path]
        if self.hide_nav_bar:
            args.extend(["--hide-nav-bar"])
        if self.hide_menu_bar:
            args.extend(["--hide-menu-bar"])
        if self.hide_status_bar:
            args.extend(["--hide-status-bar"])
        if self.fullscreen:
            args.extend(["--fullscreen"])
        if self.perfmon:
            args.extend(["--perfmon"])
        if data_plugins.is_read_only():
            args.append("--read-only")
        if self.stylesheet_path:
            args.extend(["--stylesheet", self.stylesheet_path])
        if self.home_file:
            args.extend(["--homefile", self.home_file])
        if macros is not None:
            args.extend(["-m", json.dumps(macros)])
        args.extend(["--log_level", logging.getLevelName(logging.getLogger("").getEffectiveLevel())])
        args.append(filepath)
        args.extend(self.display_args)
        args.extend(filepath_args)
        if command_line_args is not None:
            args.extend(command_line_args)
        subprocess.Popen(args, shell=False)

    def make_main_window(self, stylesheet_path=None, home_file=None, macros=None, command_line_args=None):
        """
        Instantiate a new PyDMMainWindow, add it to the application's
        list of windows. Typically, this function is only called as part
        of starting up a new process, because PyDMApplications only have
        one window per process.
        """
        main_window = PyDMMainWindow(hide_nav_bar=self.hide_nav_bar,
                                     hide_menu_bar=self.hide_menu_bar,
                                     hide_status_bar=self.hide_status_bar,
                                     home_file=home_file,
                                     macros=macros,
                                     command_line_args=command_line_args)

        self.main_window = main_window
        apply_stylesheet(stylesheet_path, widget=self.main_window)
        self.main_window.update_tools_menu()

        if self.fullscreen:
            main_window.enter_fullscreen()
        else:
            main_window.show()

    def plugin_for_channel(self, channel):
        """
        Given a PyDMChannel object, determine the appropriate plugin to use.

        Parameters
        ----------
        channel : PyDMChannel

        Returns
        -------
        PyDMPlugin
        """
        warnings.warn("'PyDMApplication.plugin_for_channel' is deprecated, "
                      "use 'pydm.data_plugins.plugin_for_address' instead.")
        if channel.address is None or channel.address == "":
            return None
        return data_plugins.plugin_for_address(channel.address)

    def add_connection(self, channel):
        """
        Add a new connection to a channel.

        Parameters
        ----------
        channel : PyDMChannel
        """
        warnings.warn("'PyDMApplication.add_connection' is deprecated, "
                      "use PyDMConnection.connect()")
        channel.connect()

    def remove_connection(self, channel):
        """
        Remove a connection to a channel.

        Parameters
        ----------
        channel : PyDMChannel
        """
        warnings.warn("'PyDMApplication.remove_connection' is deprecated, "
                      "use PyDMConnection.disconnect()")
        channel.disconnect()

    def eventFilter(self, obj, event):
        warnings.warn("'PyDMApplication.eventFilter' is deprecated, "
                      " this function is now found on PyDMWidget")
        obj.eventFilter(obj, event)

    def show_address_tooltip(self, obj, event):
        warnings.warn("'PyDMApplication.show_address_tooltip' is deprecated, "
                      " this function is now found on PyDMWidget")
        obj.show_address_tooltip(event)

    def establish_widget_connections(self, widget):
        """
        Given a widget to start from, traverse the tree of child widgets,
        and try to establish connections to any widgets with channels.

        Display subclasses which dynamically create widgets may need to
        use this method.

        Parameters
        ----------
        widget : QWidget
        """
        warnings.warn("'PyDMApplication.establish_widget_connections' is deprecated, "
                      "this function is now found on `utilities.establish_widget_connections`.")
        connection.establish_widget_connections(widget)

    def close_widget_connections(self, widget):
        """
        Given a widget to start from, traverse the tree of child widgets,
        and try to close connections to any widgets with channels.

        Parameters
        ----------
        widget : QWidget
        """
        warnings.warn(
            "'PyDMApplication.close_widget_connections' is deprecated, "
            "this function is now found on `utilities.close_widget_connections`.")
        connection.close_widget_connections(widget)



import pytest
import logging
import numpy as np
from collections import OrderedDict
from pyqtgraph import AxisItem, BarGraphItem
from unittest import mock
from ...widgets.channel import PyDMChannel
from ...widgets.scatterplot import ScatterPlotCurveItem, PyDMScatterPlot, MINIMUM_BUFFER_SIZE, DEFAULT_BUFFER_SIZE
from ...utilities import remove_protocol

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name",
    [
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, "test_name"),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, None),
        ("", "", None, None, None),
        (None, None, None, None, None),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", ScatterPlotCurveItem.REDRAW_ON_BOTH, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, "ca://test_value:Int", "test_name"),
    ]
)
def test_scatterplotcurveitem_construct(qtbot, y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name):
    plot_curve_item = ScatterPlotCurveItem(y_addr,
                                           x_addr,
                                           redraw_mode=redraw_mode,
                                           bufferSizeChannelAddress=bufferSizeChannelAddress,
                                           name=name)
    assert plot_curve_item is not None
    assert isinstance(plot_curve_item, ScatterPlotCurveItem)

    assert plot_curve_item._bufferSize == DEFAULT_BUFFER_SIZE
    assert plot_curve_item.redraw_mode == redraw_mode or ScatterPlotCurveItem.REDRAW_ON_EITHER
    assert np.array_equal(
        plot_curve_item.data_buffer,
        np.zeros((2, plot_curve_item._bufferSize), order='f', dtype=float))
    for item in "x_connected y_connected bufferSizeChannel_connected".split():
        assert getattr(plot_curve_item, item) is False
    assert plot_curve_item.points_accumulated == 0
    for item in "latest_x_value latest_y_value".split():
        assert getattr(plot_curve_item, item) is None

    keys = (
        ("x_address", x_addr),
        ("y_address", y_addr),
        ("bufferSizeChannelAddress", bufferSizeChannelAddress),
    )
    for attr, kw in keys:
        obj = getattr(plot_curve_item, attr)
        expect = kw if kw else None
        assert obj == expect, f"{attr} {kw} {expect}"


@pytest.mark.parametrize(
    "y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name",
    [
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, "test_name"),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, None),
        ("", "", None, None, None),
        (None, None, None, None, None),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", ScatterPlotCurveItem.REDRAW_ON_BOTH, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, "ca://test_value:Int", "test_name"),
    ]
)
def test_scatterplotcurveitem_to_dict(qtbot, y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name):
    plot_curve_item = ScatterPlotCurveItem(y_addr,
                                           x_addr,
                                           redraw_mode=redraw_mode,
                                           bufferSizeChannelAddress=bufferSizeChannelAddress,
                                           name=name)

    dictionary = plot_curve_item.to_dict()
    assert isinstance(dictionary, OrderedDict)

    y_name = remove_protocol(y_addr if y_addr is not None else "")
    x_name = remove_protocol(x_addr if x_addr is not None else "")
    if name is None:
        if y_addr is None and x_addr is None:
            assert dictionary["name"] == ""
        else:
            assert dictionary["name"] == f"{y_name} vs. {x_name}"
    else:
        assert dictionary["name"] == name


@pytest.mark.parametrize("new_address", [
    "new_address",
    "",
    None
])
def test_scatterplotcurveitem_properties_and_setters(qtbot, new_address):
    plot_curve_item = ScatterPlotCurveItem(new_address,
                                           new_address,
                                           bufferSizeChannelAddress=new_address)

    assert plot_curve_item.x_address in (None, new_address)
    assert plot_curve_item.y_address in (None, new_address)
    assert plot_curve_item.bufferSizeChannelAddress in (None, new_address)

    if new_address:
        assert isinstance(plot_curve_item.x_channel, PyDMChannel)
        assert plot_curve_item.x_channel.address == new_address
        assert plot_curve_item.y_channel.address == new_address
    else:
        assert plot_curve_item.x_channel is None
        assert plot_curve_item.y_channel is None
        assert plot_curve_item.bufferSizeChannel is None


def test_scatterplotcurveitem_connection_state_changed(qtbot, signals):
    plot_curve_item = ScatterPlotCurveItem(None, None)
    assert plot_curve_item.x_connected is False
    assert plot_curve_item.y_connected is False

    signals.connection_state_signal.connect(plot_curve_item.xConnectionStateChanged)
    signals.connection_state_signal.emit(True)
    assert plot_curve_item.x_connected


@pytest.mark.parametrize(
    "redraw_mode, new_data",
    [
        (ScatterPlotCurveItem.REDRAW_ON_EITHER, [(0,0), (1,1), (2,2), (2.5,3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_BOTH, [(0,0), (1,1), (2,2), (2.5,3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_X, [(0,0), (1,1), (2,2), (2.5,3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_Y, [(0,0), (1,1), (2,2), (2.5,3.1)]),
    ]
)
def test_scatterplotcurveitem_receive_values(qtbot, signals, redraw_mode, new_data):
    # REDRAW_ON_X, REDRAW_ON_Y, REDRAW_ON_EITHER, REDRAW_ON_BOTH
    plot_curve_item = ScatterPlotCurveItem(None, None, redraw_mode=redraw_mode)

    expected_data_buffer = np.zeros((2, plot_curve_item._bufferSize), order='f', dtype=float)
    expected_data_buffer[0] = plot_curve_item.data_buffer[0]
    assert np.array_equal(expected_data_buffer, plot_curve_item.data_buffer)

    # inject the new_data, point by point
    assert plot_curve_item.points_accumulated == 0
    for i, pair in enumerate(new_data):
        new_x, new_y = pair

        plot_curve_item.receiveXValue(new_x)
        assert plot_curve_item.latest_x_value == new_x
        assert plot_curve_item.points_accumulated == 2*i

        plot_curve_item.receiveYValue(new_y)
        assert plot_curve_item.latest_y_value == new_y
        assert plot_curve_item.points_accumulated == 2*i + 1


def test_scatterplotcurve_initialize_buffer(qtbot):
    plot_curve_item = ScatterPlotCurveItem(None, None)

    plot_curve_item.initialize_buffer()

    assert plot_curve_item.points_accumulated == 0
    expected_data_buffer = np.zeros((2, plot_curve_item._bufferSize), order='f', dtype=float)
    expected_data_buffer[0] = plot_curve_item.data_buffer[0]

    assert np.array_equal(expected_data_buffer, plot_curve_item.data_buffer)


@pytest.mark.parametrize("new_buffer_size, expected_set_buffer_size", [
    (0, MINIMUM_BUFFER_SIZE),
    (-5, MINIMUM_BUFFER_SIZE),
    (100, 100),
    (MINIMUM_BUFFER_SIZE + 1, MINIMUM_BUFFER_SIZE + 1)
])
def test_scatterplotcurve_get_set_reset_buffer_size(qtbot, new_buffer_size, expected_set_buffer_size):
    plot_curve_item = ScatterPlotCurveItem(None, None)

    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE

    plot_curve_item.setBufferSize(new_buffer_size)
    assert plot_curve_item.getBufferSize() == expected_set_buffer_size

    plot_curve_item.resetBufferSize()
    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE


@pytest.mark.parametrize(
    "addr, new_size, expected_buffer_size",
    [
        (None, None, DEFAULT_BUFFER_SIZE),
        ("", None, DEFAULT_BUFFER_SIZE),
        ("ca://test_value:Int", None, DEFAULT_BUFFER_SIZE),
    ]
)
def test_scatterplotcurve_get_set_reset_buffer_size(qtbot, addr, new_size, expected_buffer_size):
    plot_curve_item = ScatterPlotCurveItem(None, None, bufferSizeChannelAddress=addr)

    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE

    if new_size is not None:
        plot_curve_item.bufferSizeChannelValueReceiver(new_size)
        assert plot_curve_item.getBufferSize() == new_size
        assert plot_curve_item.data_buffer.shape == (2, new_size)
